import unittest

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression

from darts import TimeSeries
from darts.logging import get_logger
from darts.metrics import rmse
from darts.models import (
    LinearRegressionModel,
    NaiveDrift,
    NaiveSeasonal,
    RandomForest,
    RegressionEnsembleModel,
    RegressionModel,
    Theta,
)
from darts.tests.base_test_class import DartsBaseTestClass
from darts.tests.models.forecasting.test_ensemble_models import _make_ts
from darts.tests.models.forecasting.test_regression_models import train_test_split
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)

try:
    import torch

    from darts.models import BlockRNNModel, RNNModel

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not available. Some tests will be skipped.")
    TORCH_AVAILABLE = False


class RegressionEnsembleModelsTestCase(DartsBaseTestClass):
    RANDOM_SEED = 111

    sine_series = tg.sine_timeseries(
        value_frequency=(1 / 5), value_y_offset=10, length=50
    )
    lin_series = tg.linear_timeseries(length=50)

    combined = sine_series + lin_series

    seq1 = [_make_ts(0), _make_ts(10), _make_ts(20)]
    cov1 = [_make_ts(5), _make_ts(15), _make_ts(25)]

    seq2 = [_make_ts(0, 20), _make_ts(10, 20), _make_ts(20, 20)]
    cov2 = [_make_ts(5, 30), _make_ts(15, 30), _make_ts(25, 30)]

    # dummy feature and target TimeSeries instances
    ts_periodic = tg.sine_timeseries(length=500)
    ts_gaussian = tg.gaussian_timeseries(length=500)
    ts_random_walk = tg.random_walk_timeseries(length=500)

    ts_cov1 = ts_periodic.stack(ts_gaussian)
    ts_cov1 = ts_cov1.pd_dataframe()
    ts_cov1.columns = ["Periodic", "Gaussian"]
    ts_cov1 = TimeSeries.from_dataframe(ts_cov1)
    ts_sum1 = ts_periodic + ts_gaussian

    ts_cov2 = ts_sum1.stack(ts_random_walk)
    ts_sum2 = ts_sum1 + ts_random_walk

    def get_local_models(self):
        return [NaiveDrift(), NaiveSeasonal(5), NaiveSeasonal(10)]

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def get_global_models(self, output_chunk_length=5):
        return [
            RNNModel(
                input_chunk_length=20,
                output_chunk_length=output_chunk_length,
                n_epochs=1,
                random_state=42,
            ),
            BlockRNNModel(
                input_chunk_length=20,
                output_chunk_length=output_chunk_length,
                n_epochs=1,
                random_state=42,
            ),
        ]

    @staticmethod
    def get_global_ensembe_model(output_chunk_length=5):
        lags = [-1, -2, -5]
        return RegressionEnsembleModel(
            forecasting_models=[
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
            ],
            regression_train_n_points=10,
        )

    def test_accepts_different_regression_models(self):
        regr1 = LinearRegression()
        regr2 = RandomForestRegressor()
        regr3 = RandomForest(lags_future_covariates=[0])

        model0 = RegressionEnsembleModel(self.get_local_models(), 10)
        model1 = RegressionEnsembleModel(self.get_local_models(), 10, regr1)
        model2 = RegressionEnsembleModel(self.get_local_models(), 10, regr2)
        model3 = RegressionEnsembleModel(self.get_local_models(), 10, regr3)

        models = [model0, model1, model2, model3]
        for model in models:
            model.fit(series=self.combined)
            model.predict(10)

    def test_accepts_one_model(self):
        regr1 = LinearRegression()
        regr2 = RandomForest(lags_future_covariates=[0])

        model0 = RegressionEnsembleModel([self.get_local_models()[0]], 10)
        model1 = RegressionEnsembleModel([self.get_local_models()[0]], 10, regr1)
        model2 = RegressionEnsembleModel([self.get_local_models()[0]], 10, regr2)

        models = [model0, model1, model2]
        for model in models:
            model.fit(series=self.combined)
            model.predict(10)

    def test_train_n_points(self):
        regr = LinearRegressionModel(lags_future_covariates=[0])

        # same values
        _ = RegressionEnsembleModel(self.get_local_models(), 5, regr)

        # too big value to perform the split
        ensemble = RegressionEnsembleModel(self.get_local_models(), 100)
        with self.assertRaises(ValueError):
            ensemble.fit(self.combined)

        ensemble = RegressionEnsembleModel(self.get_local_models(), 50)
        with self.assertRaises(ValueError):
            ensemble.fit(self.combined)

        # too big value considering min_train_series_length
        ensemble = RegressionEnsembleModel(self.get_local_models(), 45)
        with self.assertRaises(ValueError):
            ensemble.fit(self.combined)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_torch_models_retrain(self):
        model1 = BlockRNNModel(
            input_chunk_length=12, output_chunk_length=1, random_state=0, n_epochs=2
        )
        model2 = BlockRNNModel(
            input_chunk_length=12, output_chunk_length=1, random_state=0, n_epochs=2
        )

        ensemble = RegressionEnsembleModel([model1], 5)
        ensemble.fit(self.combined)

        model1_fitted = ensemble.models[0]
        forecast1 = model1_fitted.predict(10)

        model2.fit(self.combined)
        forecast2 = model2.predict(10)

        self.assertAlmostEqual(
            sum(forecast1.values() - forecast2.values())[0], 0.0, places=2
        )

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_train_predict_global_models_univar(self):
        ensemble_models = self.get_global_models(output_chunk_length=10)
        ensemble_models.append(RegressionModel(lags=1))
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(series=self.combined)
        ensemble.predict(10)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_train_predict_global_models_multivar_no_covariates(self):
        ensemble_models = self.get_global_models(output_chunk_length=10)
        ensemble_models.append(RegressionModel(lags=1))
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(self.seq1)
        ensemble.predict(10, self.seq1)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_train_predict_global_models_multivar_with_covariates(self):
        ensemble_models = self.get_global_models(output_chunk_length=10)
        ensemble_models.append(RegressionModel(lags=1, lags_past_covariates=[-1]))
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(self.seq1, self.cov1)
        ensemble.predict(10, self.seq2, self.cov2)

    def test_train_predict_models_with_future_covariates(self):
        ensemble_models = [
            LinearRegressionModel(lags=1, lags_future_covariates=[1]),
            RandomForest(lags=1, lags_future_covariates=[1]),
        ]
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(self.sine_series, future_covariates=self.ts_cov1)
        # expected number of coefs is lags*components -> we have 1 lag for each target (1 comp)
        # and future covs (2 comp)
        expected_coefs = len(self.sine_series.components) + len(self.ts_cov1.components)
        assert len(ensemble_models[0].model.coef_) == expected_coefs
        ensemble.predict(10, self.sine_series, future_covariates=self.ts_cov1)

    def test_predict_with_target(self):
        series_long = self.combined
        series_short = series_long[:25]

        # train with a single series
        ensemble_model = self.get_global_ensembe_model()
        ensemble_model.fit(series_short, past_covariates=series_long)
        # predict after end of train series
        preds = ensemble_model.predict(n=5, past_covariates=series_long)
        self.assertTrue(isinstance(preds, TimeSeries))
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        self.assertTrue(isinstance(preds, TimeSeries))
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 2)
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 1)

        # train with multiple series
        ensemble_model = self.get_global_ensembe_model()
        ensemble_model.fit([series_short] * 2, past_covariates=[series_long] * 2)
        with self.assertRaises(ValueError):
            # predict without passing series should raise an error
            ensemble_model.predict(n=5, past_covariates=series_long)
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        self.assertTrue(isinstance(preds, TimeSeries))
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 2)
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 1)

    def helper_test_models_accuracy(
        self, model_instance, n, series, past_covariates, min_rmse
    ):
        # for every model, test whether it predicts the target with a minimum r2 score of `min_rmse`
        train_series, test_series = train_test_split(series, pd.Timestamp("20010101"))
        train_past_covariates, _ = train_test_split(
            past_covariates, pd.Timestamp("20010101")
        )

        model_instance.fit(series=train_series, past_covariates=train_past_covariates)
        prediction = model_instance.predict(n=n, past_covariates=past_covariates)
        current_rmse = rmse(test_series, prediction)

        self.assertTrue(
            current_rmse <= min_rmse,
            f"Model was not able to denoise data. A rmse score of {current_rmse} was recorded.",
        )

    def denoising_input(self):
        np.random.seed(self.RANDOM_SEED)

        ts_periodic = tg.sine_timeseries(length=500)
        ts_gaussian = tg.gaussian_timeseries(length=500)
        ts_random_walk = tg.random_walk_timeseries(length=500)

        ts_cov1 = ts_periodic.stack(ts_gaussian)
        ts_cov1 = ts_cov1.pd_dataframe()
        ts_cov1.columns = ["Periodic", "Gaussian"]
        ts_cov1 = TimeSeries.from_dataframe(ts_cov1)
        ts_sum1 = ts_periodic + ts_gaussian

        ts_cov2 = ts_sum1.stack(ts_random_walk)
        ts_sum2 = ts_sum1 + ts_random_walk

        return ts_sum1, ts_cov1, ts_sum2, ts_cov2

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_ensemble_models_denoising(self):
        # for every model, test whether it correctly denoises ts_sum using ts_gaussian and ts_sum as inputs
        # WARNING: this test isn't numerically stable, changing self.RANDOM_SEED can lead to exploding coefficients
        horizon = 10
        ts_sum1, ts_cov1, _, _ = self.denoising_input()
        torch.manual_seed(self.RANDOM_SEED)

        ensemble_models = [
            RNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
            ),
            BlockRNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
            ),
            RegressionModel(lags_past_covariates=[-1]),
        ]

        ensemble = RegressionEnsembleModel(ensemble_models, horizon)
        self.helper_test_models_accuracy(ensemble, horizon, ts_sum1, ts_cov1, 3)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_ensemble_models_denoising_multi_input(self):
        # for every model, test whether it correctly denoises ts_sum_2 using ts_random_multi and ts_sum_2 as inputs
        # WARNING: this test isn't numerically stable, changing self.RANDOM_SEED can lead to exploding coefficients
        horizon = 10
        _, _, ts_sum2, ts_cov2 = self.denoising_input()
        torch.manual_seed(self.RANDOM_SEED)

        ensemble_models = [
            RNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
            ),
            BlockRNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
            ),
            RegressionModel(lags_past_covariates=[-1]),
            RegressionModel(lags_past_covariates=[-1]),
        ]

        ensemble = RegressionEnsembleModel(ensemble_models, horizon)
        self.helper_test_models_accuracy(ensemble, horizon, ts_sum2, ts_cov2, 3)

    def test_call_backtest_regression_ensemble_local_models(self):
        series = tg.sine_timeseries(
            value_frequency=(1 / 5), value_y_offset=10, length=50
        )
        regr_train_n = 10
        ensemble = RegressionEnsembleModel(
            [NaiveSeasonal(5), Theta(2, 5)], regression_train_n_points=regr_train_n
        )
        ensemble.fit(series)
        assert max(m_.min_train_series_length for m_ in ensemble.models) == 10
        # -10 comes from the maximum minimum train series length of all models
        assert ensemble.extreme_lags == (-10 - regr_train_n, 0, None, None, None, None)
        ensemble.backtest(series)



"""
Ensemble Model Base Class
"""

from abc import abstractmethod
from functools import reduce
from typing import List, Optional, Sequence, Tuple, Union

from darts.logging import get_logger, raise_if, raise_if_not
from darts.models.forecasting.forecasting_model import (
    GlobalForecastingModel,
    LocalForecastingModel,
)
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class EnsembleModel(GlobalForecastingModel):
    """
    Abstract base class for ensemble models.
    Ensemble models take in a list of forecasting models and ensemble their predictions
    to make a single one according to the rule defined by their `ensemble()` method.

    Parameters
    ----------
    models
        List of forecasting models whose predictions to ensemble
    """

    def __init__(
        self, models: Union[List[LocalForecastingModel], List[GlobalForecastingModel]]
    ):
        raise_if_not(
            isinstance(models, list) and models,
            "Cannot instantiate EnsembleModel with an empty list of models",
            logger,
        )

        is_local_ensemble = all(
            isinstance(model, LocalForecastingModel) for model in models
        )
        self.is_global_ensemble = all(
            isinstance(model, GlobalForecastingModel) for model in models
        )

        raise_if_not(
            is_local_ensemble or self.is_global_ensemble,
            "All models must be of the same type: either GlobalForecastingModel, or LocalForecastingModel.",
            logger,
        )

        raise_if(
            any([m._fit_called for m in models]),
            "Cannot instantiate EnsembleModel with trained/fitted models. "
            "Consider resetting all models with `my_model.untrained_model()`",
            logger,
        )

        super().__init__()
        self.models = models

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ):
        """
        Fits the model on the provided series.
        Note that `EnsembleModel.fit()` does NOT call `fit()` on each of its constituent forecasting models.
        It is left to classes inheriting from EnsembleModel to do so appropriately when overriding `fit()`
        """
        raise_if(
            not self.is_global_ensemble and not isinstance(series, TimeSeries),
            "The models are of type LocalForecastingModel, which does not support training on multiple series.",
            logger,
        )
        raise_if(
            not self.is_global_ensemble and past_covariates is not None,
            "The models are of type LocalForecastingModel, which does not support past covariates.",
            logger,
        )

        is_single_series = isinstance(series, TimeSeries)

        # check that if timeseries is single series, than covariates are as well and vice versa
        error = False

        if past_covariates is not None:
            error = is_single_series != isinstance(past_covariates, TimeSeries)

        if future_covariates is not None:
            error = is_single_series != isinstance(future_covariates, TimeSeries)

        raise_if(
            error,
            "Both series and covariates have to be either univariate or multivariate.",
            logger,
        )

        super().fit(series, past_covariates, future_covariates)

        return self

    def _stack_ts_seq(self, predictions):
        # stacks list of predictions into one multivariate timeseries
        return reduce(lambda a, b: a.stack(b), predictions)

    def _stack_ts_multiseq(self, predictions_list):
        # stacks multiple sequences of timeseries elementwise
        return [self._stack_ts_seq(ts_list) for ts_list in zip(*predictions_list)]

    def _model_encoder_settings(self):
        raise NotImplementedError(
            "Encoders are not supported by EnsembleModels. Instead add encoder to the underlying `models`."
        )

    def _make_multiple_predictions(
        self,
        n: int,
        series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
    ):
        is_single_series = isinstance(series, TimeSeries) or series is None
        predictions = [
            model._predict_wrapper(
                n=n,
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
                num_samples=num_samples,
            )
            for model in self.models
        ]
        return (
            self._stack_ts_seq(predictions)
            if is_single_series
            else self._stack_ts_multiseq(predictions)
        )

    def predict(
        self,
        n: int,
        series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
        verbose: bool = False,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:

        super().predict(
            n=n,
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            num_samples=num_samples,
            verbose=verbose,
        )

        predictions = self._make_multiple_predictions(
            n=n,
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            num_samples=num_samples,
        )
        return self.ensemble(predictions, series=series)

    @abstractmethod
    def ensemble(
        self,
        predictions: Union[TimeSeries, Sequence[TimeSeries]],
        series: Optional[Sequence[TimeSeries]] = None,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """
        Defines how to ensemble the individual models' predictions to produce a single prediction.

        Parameters
        ----------
        predictions
            Individual predictions to ensemble
        series
            Sequence of timeseries to predict on. Optional, since it only makes sense for sequences of timeseries -
            local models retain timeseries for prediction.

        Returns
        -------
        TimeSeries or Sequence[TimeSeries]
            The predicted ``TimeSeries`` or sequence of ``TimeSeries`` obtained by ensembling the individual predictions
        """
        pass

    @property
    def min_train_series_length(self) -> int:
        return max(model.min_train_series_length for model in self.models)

    @property
    def min_train_samples(self) -> int:
        return max(model.min_train_samples for model in self.models)

    @property
    def extreme_lags(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
    ]:
        def find_max_lag_or_none(lag_id, aggregator) -> Optional[int]:
            max_lag = None
            for model in self.models:
                curr_lag = model.extreme_lags[lag_id]
                if max_lag is None:
                    max_lag = curr_lag
                elif curr_lag is not None:
                    max_lag = aggregator(max_lag, curr_lag)
            return max_lag

        lag_aggregators = (min, max, min, max, min, max)
        return tuple(
            find_max_lag_or_none(i, agg) for i, agg in enumerate(lag_aggregators)
        )

    def _is_probabilistic(self) -> bool:
        return all([model._is_probabilistic() for model in self.models])



"""
Additional util functions
-------------------------
"""
from enum import Enum
from functools import wraps
from inspect import Parameter, getcallargs, signature
from types import SimpleNamespace
from typing import Callable, Iterator, List, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm
from tqdm.notebook import tqdm as tqdm_notebook

from darts import TimeSeries
from darts.logging import get_logger, raise_if_not, raise_log
from darts.utils.timeseries_generation import generate_index

try:
    from IPython import get_ipython
except ModuleNotFoundError:
    get_ipython = None

logger = get_logger(__name__)


# Enums
class SeasonalityMode(Enum):
    MULTIPLICATIVE = "multiplicative"
    ADDITIVE = "additive"
    NONE = None


class TrendMode(Enum):
    LINEAR = "linear"
    EXPONENTIAL = "exponential"


class ModelMode(Enum):
    MULTIPLICATIVE = "multiplicative"
    ADDITIVE = "additive"
    NONE = None


# TODO: we do not check the time index here
def retain_period_common_to_all(series: List[TimeSeries]) -> List[TimeSeries]:
    """
    Trims all series in the provided list, if necessary, so that the returned time series have
    a common span (corresponding to largest time sub-interval common to all series).

    Parameters
    ----------
    series
        The list of series to consider.

    Raises
    ------
    ValueError
        If no common time sub-interval exists

    Returns
    -------
    List[TimeSeries]
        A list of series, where each series have the same span
    """

    last_first = max(map(lambda s: s.start_time(), series))
    first_last = min(map(lambda s: s.end_time(), series))

    if last_first >= first_last:
        raise_log(
            ValueError("The provided time series must have nonzero overlap"), logger
        )

    return list(map(lambda s: s.slice(last_first, first_last), series))


def _build_tqdm_iterator(iterable, verbose, **kwargs):
    """
    Build an iterable, possibly using tqdm (either in notebook or regular mode)

    Parameters
    ----------
    iterable
    verbose
    total
        Length of the iterator, helps in cases where tqdm is not detecting the total length.

    Returns
    -------
    """

    def _isnotebook():
        if get_ipython is None:
            return False
        try:
            shell = get_ipython().__class__.__name__
            if shell == "ZMQInteractiveShell":
                return True  # Jupyter notebook or qtconsole
            elif shell == "TerminalInteractiveShell":
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter

    if verbose:
        if _isnotebook():
            iterator = tqdm_notebook(iterable, **kwargs)
        else:
            iterator = tqdm(iterable, **kwargs)

    else:
        iterator = iterable
    return iterator


# Types for sanity checks decorator
A = TypeVar("A")
B = TypeVar("B")
T = TypeVar("T")


def _with_sanity_checks(
    *sanity_check_methods: str,
) -> Callable[[Callable[[A, B], T]], Callable[[A, B], T]]:
    """
    Decorator allowing to specify some sanity check method(s) to be used on a class method.
    The decorator guarantees that args and kwargs from the method to sanitize will be available in the
    sanity check methods as specified in the sanitized method's signature, irrespective of how it was called.

    Parameters
    ----------
    *sanity_check_methods
        one or more sanity check methods that will be called with all the parameter of the decorated method.

    Returns
    -------
    A Callable corresponding to the decorated method.

    Examples
    --------
    class Model:
        def _a_sanity_check(self, *args, **kwargs):
            raise_if_not(kwargs['b'] == kwargs['c'], 'b must equal c', logger)
        @_with_sanity_checks("_a_sanity_check")
        def fit(self, a, b=0, c=0):
            # at this point we can safely assume that 'b' and 'c' are equal...
            ...
    """

    def decorator(method_to_sanitize: Callable[[A, B], T]) -> Callable[[A, B], T]:
        @wraps(method_to_sanitize)
        def sanitized_method(self, *args: A, **kwargs: B) -> T:
            for sanity_check_method in sanity_check_methods:
                # Convert all arguments into keyword arguments
                all_as_kwargs = getcallargs(method_to_sanitize, self, *args, **kwargs)

                # Then separate args from kwargs according to the function's signature
                only_args = all_as_kwargs.copy()
                only_kwargs = all_as_kwargs.copy()

                for param_name, param in signature(
                    method_to_sanitize
                ).parameters.items():
                    if (
                        param.default == Parameter.empty
                        and param.kind != Parameter.VAR_POSITIONAL
                    ):
                        only_kwargs.pop(param_name)
                    else:
                        only_args.pop(param_name)

                only_args.pop("self")

                getattr(self, sanity_check_method)(*only_args.values(), **only_kwargs)
            return method_to_sanitize(self, *only_args.values(), **only_kwargs)

        return sanitized_method

    return decorator


def _historical_forecasts_general_checks(series, kwargs):
    """
    Performs checks common to ForecastingModel and RegressionModel backtest() methods
    Parameters
    ----------
    series
        Either series when called from ForecastingModel, or target_series if called from RegressionModel
    signature_params
        A dictionary of the signature parameters of the calling method, to get the default values
        Typically would be signature(self.backtest).parameters
    kwargs
        Params specified by the caller of backtest(), they take precedence over the arguments' default values
    """

    # parse kwargs
    n = SimpleNamespace(**kwargs)

    # check forecast horizon
    raise_if_not(
        n.forecast_horizon > 0,
        "The provided forecasting horizon must be a positive integer.",
        logger,
    )

    # check stride
    raise_if_not(
        n.stride > 0,
        "The provided stride parameter must be a positive integer.",
        logger,
    )

    series = series2seq(series)

    if n.start is not None:
        # check start parameter in general (non series dependent)
        if not isinstance(n.start, (float, int, np.int64, pd.Timestamp)):
            raise_log(
                TypeError(
                    "`start` needs to be either `float`, `int`, `pd.Timestamp` or `None`"
                ),
                logger,
            )
        if isinstance(n.start, float):
            raise_if_not(
                0.0 <= n.start <= 1.0, "`start` should be between 0.0 and 1.0.", logger
            )
        elif isinstance(n.start, (int, np.int64)):
            raise_if_not(
                n.start >= 0, "if `start` is an integer, must be `>= 0`.", logger
            )

        # verbose error messages
        if not isinstance(n.start, pd.Timestamp):
            start_value_msg = f"`start` value `{n.start}` corresponding to timestamp"
        else:
            start_value_msg = "`start` time"
        for idx, series_ in enumerate(series):
            # check specifically for int and Timestamp as error by `get_timestamp_at_point` is too generic
            if isinstance(n.start, pd.Timestamp):
                if n.start > series_.end_time():
                    raise_log(
                        ValueError(
                            f"`start` time `{n.start}` is after the last timestamp `{series_.end_time()}` of the "
                            f"series at index: {idx}."
                        ),
                        logger,
                    )
                elif n.start < series_.start_time():
                    raise_log(
                        ValueError(
                            f"`start` time `{n.start}` is before the first timestamp `{series_.start_time()}` of the "
                            f"series at index: {idx}."
                        ),
                        logger,
                    )
            elif isinstance(n.start, (int, np.int64)):
                if (
                    series_.has_datetime_index
                    or (series_.has_range_index and series_.freq == 1)
                ) and n.start >= len(series_):
                    raise_log(
                        ValueError(
                            f"`start` index `{n.start}` is out of bounds for series of length {len(series_)} "
                            f"at index: {idx}."
                        ),
                        logger,
                    )
                elif (
                    series_.has_range_index and series_.freq > 1
                ) and n.start > series_.time_index[-1]:
                    raise_log(
                        ValueError(
                            f"`start` index `{n.start}` is larger than the last index `{series_.time_index[-1]}` "
                            f"for series at index: {idx}."
                        ),
                        logger,
                    )

            start = series_.get_timestamp_at_point(n.start)
            if n.retrain is not False and start == series_.start_time():
                raise_log(
                    ValueError(
                        f"{start_value_msg} `{start}` is the first timestamp of the series {idx}, resulting in an "
                        f"empty training set."
                    ),
                    logger,
                )

            # check that overlap_end and start together form a valid combination
            overlap_end = n.overlap_end
            if not overlap_end and not (
                start + (series_.freq * (n.forecast_horizon - 1)) in series_
            ):
                raise_log(
                    ValueError(
                        f"{start_value_msg} `{start}` is too late in the series {idx} to make any predictions with "
                        f"`overlap_end` set to `False`."
                    ),
                    logger,
                )


def _parallel_apply(
    iterator: Iterator[Tuple], fn: Callable, n_jobs: int, fn_args, fn_kwargs
) -> List:
    """
    Utility function that parallelise the execution of a function over an Iterator

    Parameters
    ----------
    iterator (Iterator[Tuple])
        Iterator which returns tuples of input value to feed to fn. Constant `args` and `kwargs` should passed through
        `fn_args` and  `fn_kwargs` respectively.
    fn (Callable)
        The function to be parallelized.
    n_jobs (int)
        The number of jobs to run in parallel. Defaults to `1` (sequential). Setting the parameter to `-1` means using
        all the available processors.
        Note: for a small amount of data, the parallelisation overhead could end up increasing the total
        required amount of time.
    fn_args
        Additional arguments for each `fn()` call
    fn_kwargs
        Additional keyword arguments for each `fn()` call

    """

    returned_data = Parallel(n_jobs=n_jobs)(
        delayed(fn)(*sample, *fn_args, **fn_kwargs) for sample in iterator
    )
    return returned_data


def _check_quantiles(quantiles):
    raise_if_not(
        all([0 < q < 1 for q in quantiles]),
        "All provided quantiles must be between 0 and 1.",
    )

    # we require the median to be present and the quantiles to be symmetric around it,
    # for correctness of sampling.
    median_q = 0.5
    raise_if_not(
        median_q in quantiles, "median quantile `q=0.5` must be in `quantiles`"
    )
    is_centered = [
        -1e-6 < (median_q - left_q) + (median_q - right_q) < 1e-6
        for left_q, right_q in zip(quantiles, quantiles[::-1])
    ]
    raise_if_not(
        all(is_centered),
        "quantiles lower than `q=0.5` need to share same difference to `0.5` as quantiles "
        "higher than `q=0.5`",
    )


def series2seq(
    ts: Optional[Union[TimeSeries, Sequence[TimeSeries]]]
) -> Optional[Sequence[TimeSeries]]:
    """If `ts` is a single TimeSeries, return it as a list of a single TimeSeries.

    Parameters
    ----------
    ts
        None, a single TimeSeries, or a sequence of TimeSeries

    Returns
    -------
        `ts` if `ts` is not a TimeSeries, else `[ts]`

    """
    return [ts] if isinstance(ts, TimeSeries) else ts


def seq2series(
    ts: Optional[Union[TimeSeries, Sequence[TimeSeries]]]
) -> Optional[TimeSeries]:
    """If `ts` is a Sequence with only a single series, return the single series as TimeSeries.

    Parameters
    ----------
    ts
        None, a single TimeSeries, or a sequence of TimeSeries

    Returns
    -------
        `ts` if `ts` if is not a single element TimeSeries sequence, else `ts[0]`

    """

    return ts[0] if isinstance(ts, Sequence) and len(ts) == 1 else ts


def slice_index(
    index: Union[pd.RangeIndex, pd.DatetimeIndex],
    start: Union[int, pd.Timestamp],
    end: Union[int, pd.Timestamp],
) -> Union[pd.RangeIndex, pd.DatetimeIndex]:
    """
    Returns a new Index with the same type as the input `index`, containing the values between `start`
    and `end` included. If start and end are not in the index, the closest values are used instead.
    The start and end values can be either integers (in which case they are interpreted as indices),
    or pd.Timestamps (in which case they are interpreted as actual timestamps).


    Parameters
    ----------
    index
        The index to slice.
    start
        The start of the returned index.
    end
        The end of the returned index.

    Returns
    -------
    Union[pd.RangeIndex, pd.DatetimeIndex]
        A new index with the same type as the input `index`, but with only the values between `start` and `end`
        included.
    """

    if type(start) != type(end):
        raise_log(
            ValueError(
                "start and end values must be of the same type (either both integers or both pd.Timestamps)"
            ),
            logger,
        )

    if isinstance(start, pd.Timestamp) and isinstance(index, pd.RangeIndex):
        raise_log(
            ValueError(
                "start and end values are a pd.Timestamp, but time_index is a RangeIndex. "
                "Please provide an integer start value."
            ),
            logger,
        )
    if isinstance(start, int) and isinstance(index, pd.DatetimeIndex):
        raise_log(
            ValueError(
                "start and end value are integer, but time_index is a RangeIndex. "
                "Please provide an integer end value."
            ),
            logger,
        )

    start_idx = index.get_indexer(generate_index(start, length=1), method="nearest")[0]
    end_idx = index.get_indexer(generate_index(end, length=1), method="nearest")[0]

    return index[start_idx : end_idx + 1]


def drop_before_index(
    index: Union[pd.RangeIndex, pd.DatetimeIndex],
    split_point: Union[int, pd.Timestamp],
) -> Union[pd.RangeIndex, pd.DatetimeIndex]:
    """
    Drops everything before the provided time `split_point` (excluded) from the index.

    Parameters
    ----------
    index
        The index to drop values from.
    split_point
        The timestamp that indicates cut-off time.

    Returns
    -------
    Union[pd.RangeIndex, pd.DatetimeIndex]
        A new index with values before `split_point` dropped.
    """
    return slice_index(index, split_point, index[-1])


def drop_after_index(
    index: Union[pd.RangeIndex, pd.DatetimeIndex],
    split_point: Union[int, pd.Timestamp],
) -> Union[pd.RangeIndex, pd.DatetimeIndex]:
    """
    Drops everything after the provided time `split_point` (excluded) from the index.

    Parameters
    ----------
    index
        The index to drop values from.
    split_point
        The timestamp that indicates cut-off time.

    Returns
    -------
    Union[pd.RangeIndex, pd.DatetimeIndex]
        A new index with values after `split_point` dropped.
    """

    return slice_index(index, index[0], split_point)


def get_single_series(
    ts: Optional[Union[TimeSeries, Sequence[TimeSeries]]]
) -> Optional[TimeSeries]:
    """Returns a single (first) TimeSeries or `None` from `ts`. Returns `ts` if  `ts` is a TimeSeries, `ts[0]` if
    `ts` is a Sequence of TimeSeries. Otherwise, returns `None`.

    Parameters
    ----------
    ts
        None, a single TimeSeries, or a sequence of TimeSeries.

    Returns
    -------
        `ts` if  `ts` is a TimeSeries, `ts[0]` if `ts` is a Sequence of TimeSeries. Otherwise, returns `None`

    """
    if isinstance(ts, TimeSeries) or ts is None:
        return ts
    else:
        return ts[0]



"""
Exponential Smoothing
---------------------
"""

from typing import Optional

import numpy as np
import statsmodels.tsa.holtwinters as hw

from darts.logging import get_logger
from darts.models.forecasting.forecasting_model import LocalForecastingModel
from darts.timeseries import TimeSeries
from darts.utils.utils import ModelMode, SeasonalityMode

logger = get_logger(__name__)


class ExponentialSmoothing(LocalForecastingModel):
    def __init__(
        self,
        trend: Optional[ModelMode] = ModelMode.ADDITIVE,
        damped: Optional[bool] = False,
        seasonal: Optional[SeasonalityMode] = SeasonalityMode.ADDITIVE,
        seasonal_periods: Optional[int] = None,
        random_state: int = 0,
        **fit_kwargs,
    ):

        """Exponential Smoothing

        This is a wrapper around
        `statsmodels  Holt-Winters' Exponential Smoothing
        <https://www.statsmodels.org/stable/generated/statsmodels.tsa.holtwinters.ExponentialSmoothing.html>`_;
        we refer to this link for the original and more complete documentation of the parameters.

        `trend` must be a ``ModelMode`` Enum member. You can access the Enum with
         ``from darts.utils.utils import ModelMode``.
        `seasonal` must be a ``SeasonalityMode`` Enum member. You can access the Enum with
        ``from darts.utils.utils import SeasonalityMode``.

        ``ExponentialSmoothing(trend=ModelMode.NONE, seasonal=SeasonalityMode.NONE)`` corresponds to a single
        exponential smoothing.
        ``ExponentialSmoothing(trend=ModelMode.ADDITIVE, seasonal=SeasonalityMode.NONE)`` corresponds to a Holt's
        exponential smoothing.

        Please note that automatic `seasonal_period` selection (setting the `seasonal_periods` parameter equal to
        `None`) can sometimes lead to errors if the input time series is too short. In these cases we suggest to
        manually set the `seasonal_periods` parameter to a positive integer.

        Parameters
        ----------
        trend
            Type of trend component. Either ``ModelMode.ADDITIVE``, ``ModelMode.MULTIPLICATIVE``, ``ModelMode.NONE``,
            or ``None``. Defaults to ``ModelMode.ADDITIVE``.
        damped
            Should the trend component be damped. Defaults to False.
        seasonal
            Type of seasonal component. Either ``SeasonalityMode.ADDITIVE``, ``SeasonalityMode.MULTIPLICATIVE``,
            ``SeasonalityMode.NONE``, or ``None``. Defaults to ``SeasonalityMode.ADDITIVE``.
        seasonal_periods
            The number of periods in a complete seasonal cycle, e.g., 4 for quarterly data or 7 for daily
            data with a weekly cycle. If not set, inferred from frequency of the series.
        fit_kwargs
            Some optional keyword arguments that will be used to call
            :func:`statsmodels.tsa.holtwinters.ExponentialSmoothing.fit()`.
            See `the documentation
            <https://www.statsmodels.org/stable/generated/statsmodels.tsa.holtwinters.ExponentialSmoothing.fit.html>`_.
        """
        super().__init__()
        self.trend = trend
        self.damped = damped
        self.seasonal = seasonal
        self.infer_seasonal_periods = seasonal_periods is None
        self.seasonal_periods = seasonal_periods
        self.fit_kwargs = fit_kwargs
        self.model = None
        np.random.seed(random_state)

    def fit(self, series: TimeSeries):
        super().fit(series)
        self._assert_univariate(series)
        series = self.training_series

        # if the model was initially created with `self.seasonal_periods=None`, make sure that
        # the model will try to automatically infer the index, otherwise it should use the
        # provided `seasonal_periods` value
        seasonal_periods_param = (
            None if self.infer_seasonal_periods else self.seasonal_periods
        )

        # set the seasonal periods parameter to a default value if it was not provided explicitly
        # and if it cannot be inferred due to the lack of a datetime index
        if self.seasonal_periods is None and series.has_range_index:
            seasonal_periods_param = 12

        hw_model = hw.ExponentialSmoothing(
            series.values(copy=False),
            trend=self.trend if self.trend is None else self.trend.value,
            damped_trend=self.damped,
            seasonal=self.seasonal if self.seasonal is None else self.seasonal.value,
            seasonal_periods=seasonal_periods_param,
            freq=series.freq if series.has_datetime_index else None,
            dates=series.time_index if series.has_datetime_index else None,
        )
        hw_results = hw_model.fit(**self.fit_kwargs)
        self.model = hw_results

        if self.infer_seasonal_periods:
            self.seasonal_periods = hw_model.seasonal_periods

        return self

    def predict(self, n, num_samples=1, verbose: bool = False):
        super().predict(n, num_samples)

        if num_samples == 1:
            forecast = self.model.forecast(n)
        else:
            forecast = np.expand_dims(
                self.model.simulate(n, repetitions=num_samples), axis=1
            )

        return self._build_forecast_series(forecast)

    def _is_probabilistic(self) -> bool:
        return True

    @property
    def min_train_series_length(self) -> int:
        if self.seasonal_periods is not None and self.seasonal_periods > 1:
            return 2 * self.seasonal_periods
        return 3



"""
Anomaly Detection
-----------------

A suite of tools for performing anomaly detection and classification
on time series.

* `Anomaly Scorers <https://unit8co.github.io/darts/generated_api/darts.ad.scorers.html>`_
  are at the core of the anomaly detection module. They
  produce anomaly scores time series, either for single series (``score()``),
  or for series accompanied by some predictions (``score_from_prediction()``).
  Scorers can be trainable (e.g., ``KMeansScorer``) or not (e.g., ``NormScorer``).

* `Anomaly Models <https://unit8co.github.io/darts/generated_api/darts.ad.anomaly_model.html>`_ 
  offer a convenient way to produce anomaly scores from any of Darts
  forecasting models (``ForecastingAnomalyModel``) or filtering models (``FilteringAnomalyModel``),
  by comparing models' predictions with actual observations.
  These classes take as parameters one Darts model, and one or multiple scorers, and can be readily
  used to produce anomaly scores with the ``score()`` method.

* `Anomaly Detectors <https://unit8co.github.io/darts/generated_api/darts.ad.detectors.html>`_: 
  transform raw time series (such as anaomly scores) into binary anomaly time series.

* `Anomaly Aggregators <https://unit8co.github.io/darts/generated_api/darts.ad.aggregators.html>`_:
  combine multiple binary anomaly time series (in the form of multivariate time series)
  into a single binary anomaly time series applying boolean logic.
"""

# anomaly aggregators
from .aggregators import AndAggregator, EnsembleSklearnAggregator, OrAggregator

# anomaly models
from .anomaly_model import FilteringAnomalyModel, ForecastingAnomalyModel

# anomaly detectors
from .detectors import QuantileDetector, ThresholdDetector

# anomaly scorers
from .scorers import (
    CauchyNLLScorer,
    DifferenceScorer,
    ExponentialNLLScorer,
    GammaNLLScorer,
    GaussianNLLScorer,
    KMeansScorer,
    LaplaceNLLScorer,
    NormScorer,
    PoissonNLLScorer,
    PyODScorer,
    WassersteinScorer,
)



"""
Detector Base Classes
"""

# TODO:
#     - check error message and add name of variable in the message error
#     - rethink the positionning of fun _check_param()
#     - add possibility to input a list of param rather than only one number
#     - add more complex detectors
#         - create an ensemble fittable detector

from abc import ABC, abstractmethod
from typing import Any, Sequence, Union

from darts import TimeSeries
from darts.ad.utils import eval_accuracy_from_binary_prediction
from darts.logging import raise_if_not


class Detector(ABC):
    """Base class for all detectors"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def detect(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Detect anomalies on given time series.

        Parameters
        ----------
        series
            series on which to detect anomalies.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            binary prediciton (1 if considered as an anomaly, 0 if not)
        """

        list_series = [series] if not isinstance(series, Sequence) else series

        raise_if_not(
            all([isinstance(s, TimeSeries) for s in list_series]),
            "all series in `series` must be of type TimeSeries.",
        )

        raise_if_not(
            all([s.is_deterministic for s in list_series]),
            "all series in `series` must be deterministic (number of samples equal to 1).",
        )

        detected_series = []
        for s in list_series:
            detected_series.append(self._detect_core(s))

        if len(detected_series) == 1 and not isinstance(series, Sequence):
            return detected_series[0]
        else:
            return detected_series

    @abstractmethod
    def _detect_core(self, input: Any) -> Any:
        pass

    def eval_accuracy(
        self,
        actual_anomalies: Union[TimeSeries, Sequence[TimeSeries]],
        anomaly_score: Union[TimeSeries, Sequence[TimeSeries]],
        window: int = 1,
        metric: str = "recall",
    ) -> Union[float, Sequence[float], Sequence[Sequence[float]]]:
        """Score the results against true anomalies.

        Parameters
        ----------
        actual_anomalies
            The ground truth of the anomalies (1 if it is an anomaly and 0 if not).
        anomaly_score
            Series indicating how anomoulous each window of size w is.
        window
            Integer value indicating the number of past samples each point represents
            in the anomaly_score.
        metric
            Metric function to use. Must be one of "recall", "precision",
            "f1", and "accuracy".
            Default: "recall"

        Returns
        -------
        Union[float, Sequence[float], Sequence[Sequence[float]]]
            Metric results for each anomaly score
        """

        if isinstance(anomaly_score, Sequence):
            raise_if_not(
                all([isinstance(s, TimeSeries) for s in anomaly_score]),
                "all series in `anomaly_score` must be of type TimeSeries.",
            )

            raise_if_not(
                all([s.is_deterministic for s in anomaly_score]),
                "all series in `anomaly_score` must be deterministic (number of samples equal to 1).",
            )
        else:
            raise_if_not(
                isinstance(anomaly_score, TimeSeries),
                f"Input `anomaly_score` must be of type TimeSeries, found {type(anomaly_score)}.",
            )

            raise_if_not(
                anomaly_score.is_deterministic,
                "Input `anomaly_score` must be deterministic (number of samples equal to 1).",
            )

        return eval_accuracy_from_binary_prediction(
            actual_anomalies, self.detect(anomaly_score), window, metric
        )


class FittableDetector(Detector):
    """Base class of Detectors that need training."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._fit_called = False

    def detect(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Detect anomalies on given time series.

        Parameters
        ----------
        series
            series on which to detect anomalies.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            binary prediciton (1 if considered as an anomaly, 0 if not)
        """

        list_series = [series] if not isinstance(series, Sequence) else series

        raise_if_not(
            self._fit_called,
            "The Detector has not been fitted yet. Call `fit()` first.",
        )

        raise_if_not(
            all([self.width_trained_on == s.width for s in list_series]),
            "all series in `series` must have the same number of components as the data "
            + "used for training the detector model, number of components in training: "
            + f" {self.width_trained_on}.",
        )

        return super().detect(series)

    @abstractmethod
    def _fit_core(self, input: Any) -> Any:
        pass

    def fit(self, series: Union[TimeSeries, Sequence[TimeSeries]]) -> None:
        """Trains the detector on the given time series.

        Parameters
        ----------
        series
            Time series to be used to train the detector.

        Returns
        -------
        self
            Fitted Detector.
        """

        list_series = [series] if not isinstance(series, Sequence) else series

        raise_if_not(
            all([isinstance(s, TimeSeries) for s in list_series]),
            "all series in `series` must be of type TimeSeries.",
        )

        raise_if_not(
            all([s.is_deterministic for s in list_series]),
            "all series in `series` must be deterministic (number of samples equal to 1).",
        )

        self.width_trained_on = list_series[0].width

        raise_if_not(
            all([s.width == self.width_trained_on for s in list_series]),
            "all series in `series` must have the same number of components.",
        )

        self._fit_called = True
        return self._fit_core(list_series)

    def fit_detect(
        self, series: Union[TimeSeries, Sequence[TimeSeries]]
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Trains the detector and detects anomalies on the same series.

        Parameters
        ----------
        series
            Time series to be used for training and be detected for anomalies.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            Binary prediciton (1 if considered as an anomaly, 0 if not)
        """
        self.fit(series)
        return self.detect(series)



import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.datasets import AirPassengersDataset
from darts.metrics import mae
from darts.models import LinearRegressionModel, StatsForecastAutoETS
from darts.tests.base_test_class import DartsBaseTestClass


class StatsForecastAutoETSTestCase(DartsBaseTestClass):
    # real timeseries for functionality tests
    ts_passengers = AirPassengersDataset().load()
    ts_pass_train, ts_pass_val = ts_passengers.split_after(pd.Timestamp("19570101"))

    # as future covariates we want a trend
    trend_values = np.arange(start=1, stop=len(ts_passengers) + 1)
    trend_times = ts_passengers.time_index
    ts_trend = TimeSeries.from_times_and_values(
        times=trend_times, values=trend_values, columns=["trend"]
    )
    ts_trend_train, ts_trend_val = ts_trend.split_after(pd.Timestamp("19570101"))

    def test_fit_on_residuals(self):
        model = StatsForecastAutoETS(season_length=12, model="ZZZ")

        # test if we are indeed fitting the AutoETS on the residuals of the linear regression
        model.fit(series=self.ts_pass_train, future_covariates=self.ts_trend_train)

        # create the residuals from the linear regression
        fitted_values_linreg = model._linreg.model.predict(
            X=self.ts_trend_train.values(copy=False)
        )
        fitted_values_linreg_ts = TimeSeries.from_times_and_values(
            times=self.ts_pass_train.time_index, values=fitted_values_linreg
        )
        resids = self.ts_pass_train - fitted_values_linreg_ts

        # now make in-sample predictions with the AutoETS model
        in_sample_preds = model.model.predict_in_sample()["fitted"]
        ts_in_sample_preds = TimeSeries.from_times_and_values(
            times=self.ts_pass_train.time_index, values=in_sample_preds
        )

        # compare in-sample predictions to the residuals they have supposedly been fitted on
        current_mae = mae(resids, ts_in_sample_preds)

        self.assertTrue(current_mae < 9)

    def test_fit_a_linreg(self):
        model = StatsForecastAutoETS(season_length=12, model="ZZZ")
        model.fit(series=self.ts_pass_train, future_covariates=self.ts_trend_train)

        # check if linear regression was fit
        self.assertIsNotNone(model._linreg)
        self.assertTrue(model._linreg._fit_called)

        # fit a linear regression
        linreg = LinearRegressionModel(lags_future_covariates=[0])
        linreg.fit(series=self.ts_pass_train, future_covariates=self.ts_trend_train)

        # check if the linear regression was fit on the same data by checking if the coefficients are equal
        self.assertEqual(model._linreg.model.coef_, linreg.model.coef_)



"""
Encoder Base Classes
--------------------
"""

from abc import ABC, abstractmethod
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.dataprocessing.transformers import FittableDataTransformer
from darts.logging import get_logger, raise_if, raise_log
from darts.utils.timeseries_generation import generate_index

SupportedIndex = Union[pd.DatetimeIndex, pd.RangeIndex]
EncoderOutputType = Optional[Union[Sequence[TimeSeries], List[TimeSeries]]]
logger = get_logger(__name__)


class CovariatesIndexGenerator(ABC):
    def __init__(
        self,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[List[int]] = None,
    ):
        """:class:`CovariatesIndexGenerator` generates a time index for covariates at training and inference /
        prediction time with methods :func:`generate_train_idx()`, and :func:`generate_inference_idx()`.
        Without user `covariates`, it generates the minimum required covariate times spans for the corresponding
        scenarios described below. With user `covariates`, it simply copies and returns the `covariates` time index.

        It can be used:
        A   in combination with :class:`LocalForecastingModel`, or in a model agnostic scenario:
                All parameters can be ignored. This scenario is only supported by
                :class:`FutureCovariatesIndexGenerator`.
        B   in combination with :class:`RegressionModel`:
                Set `input_chunk_length`, `output_chunk_length`, and `lags_covariates`.
                `input_chunk_length` is the absolute value of the minimum target lag `abs(min(lags))` used with the
                regression model.
                Set `output_chunk_length`, and `lags_covariates` with the identical values used at forecasting model
                creation. For the covariates lags, use `lags_past_covariates` for class:`PastCovariatesIndexGenerator`,
                and `lags_future_covariates` for class:`PastCovariatesIndexGenerator`.
        C   in combination with :class:`TorchForecastingModel`:
                Set `input_chunk_length`, and `output_chunk_length` with the identical values used at forecasting model
                creation.

        Parameters
        ----------
        input_chunk_length
            Optionally, the number of input target time steps per chunk. Only required in scenarios B, C.
            Corresponds to `input_chunk_length` for :class:`TorchForecastingModel`, or to the absolute minimum target
            lag value `abs(min(lags))` for :class:`RegressionModel`.
        output_chunk_length
            Optionally, the number of output target time steps per chunk. Only required in scenarios B, and C.
            Corresponds to `output_chunk_length` for both :class:`TorchForecastingModel`, and :class:`RegressionModel`.
        lags_covariates
            Optionally, a list of integers giving the covariates lags used for Darts' RegressionModels. Only required
            in scenario B. Corresponds to the lag values from `lags_past_covariates` for past covariates, and
            `lags_future_covariates` for future covariates.
        """
        # check that parameters match one of the scenarios
        self._verify_scenario(input_chunk_length, output_chunk_length, lags_covariates)

        # input/output chunk length are guaranteed to both be `None`, or both be defined
        self.input_chunk_length = input_chunk_length
        self.output_chunk_length = output_chunk_length

        # check lags validity
        min_covariates_lag = (
            min(lags_covariates) if lags_covariates is not None else None
        )
        max_covariates_lag = (
            max(lags_covariates) if lags_covariates is not None else None
        )
        self._verify_lags(min_covariates_lag, max_covariates_lag)

        # from verification min/max lags are guaranteed to either both be None, or both be an integer
        if min_covariates_lag is not None:
            # we add 1 to the lags so that shift == 0 represents the end of the target series (forecasting point)
            shift_start = min_covariates_lag + 1
            shift_end = max_covariates_lag + 1
        else:
            shift_start = None
            shift_end = None
        self.shift_start = shift_start
        self.shift_end = shift_end

    @abstractmethod
    def generate_train_idx(
        self, target: TimeSeries, covariates: Optional[TimeSeries] = None
    ) -> Tuple[SupportedIndex, pd.Timestamp]:
        """
        Generates/extracts time index (or integer index) for covariates at model training time.

        Parameters
        ----------
        target
            The target TimeSeries used during training.
        covariates
            Optionally, the covariates used for training.
            If given, the returned time index is equal to the `covariates` time index. Else, the returned time index
            covers the minimum required covariate time span for training a specific forecasting model. These
            requirements are derived from parameters set at :class:`CovariatesIndexGenerator` creation.
        """
        pass

    @abstractmethod
    def generate_inference_idx(
        self, n: int, target: TimeSeries, covariates: Optional[TimeSeries] = None
    ) -> Tuple[SupportedIndex, pd.Timestamp]:
        """
        Generates/extracts time index (or integer index) for covariates at model inference / prediction time.

        Parameters
        ----------
        n
            The forecasting horizon.
        target
            The target TimeSeries used during training or passed to prediction as `series`.
        covariates
            Optionally, the covariates used for prediction.
            If given, the returned time index is equal to the `covariates` time index. Else, the returned time index
            covers the minimum required covariate time spans for performing inference / prediction with a specific
            forecasting model. These requirements are derived from parameters set at :class:`CovariatesIndexGenerator`
            creation.
        """
        pass

    @property
    @abstractmethod
    def base_component_name(self) -> str:
        """Returns the index generator base component name.
        - "pc": past covariates
        - "fc": future covariates
        """
        pass

    def _verify_scenario(
        self,
        input_chunk_length: Optional[int] = None,
        output_chunk_length: Optional[int] = None,
        lags_covariates: Optional[List[int]] = None,
    ):
        # LocalForecastingModel, or model agnostic (only supported by future covariates)
        is_scenario_a = (
            isinstance(self, FutureCovariatesIndexGenerator)
            and input_chunk_length is None
            and output_chunk_length is None
            and lags_covariates is None
        )
        # RegressionModel
        is_scenario_b = (
            input_chunk_length is not None
            and output_chunk_length is not None
            and lags_covariates is not None
        )
        # TorchForecastingModel
        is_scenario_c = (
            input_chunk_length is not None
            and output_chunk_length is not None
            and lags_covariates is None
        )

        if not any([is_scenario_a, is_scenario_b, is_scenario_c]):
            raise_log(
                ValueError(
                    "Invalid `CovariatesIndexGenerator` parameter combination: Could not be mapped to an existing "
                    "scenario, as defined in "
                    "https://unit8co.github.io/darts/generated_api/darts.dataprocessing.encoders.encoder_base.html"
                    "#darts.dataprocessing.encoders.encoder_base.CovariatesIndexGenerator"
                ),
                logger=logger,
            )

    def _verify_lags(self, min_covariates_lag, max_covariates_lag):
        """Check the base requirements for `min_covariates_lag` and `max_covariates_lag`:
        - both must either be None or an integer
        - min_covariates_lag < max_covariates_lag

        This method can be extended by subclasses for past and future covariates lag requirements.
        """
        # check that either None one of min/max_covariates_lag are given, or both are given
        if (min_covariates_lag is not None and max_covariates_lag is None) or (
            min_covariates_lag is None and max_covariates_lag is not None
        ):
            raise_log(
                ValueError(
                    "`min_covariates_lag` and `max_covariates_lag` must either both be `None` or both be integers"
                ),
                logger=logger,
            )
        if min_covariates_lag is not None:
            # check that if one of the two is given, both must be integers
            if not isinstance(min_covariates_lag, int) or not isinstance(
                max_covariates_lag, int
            ):
                raise_log(
                    ValueError(
                        "`min_covariates_lag` and `max_covariates_lag` must be both be integers."
                    ),
                    logger=logger,
                )
            # minimum lag must be less than maximum lag
            if min_covariates_lag > max_covariates_lag:
                raise_log(
                    ValueError(
                        "`min_covariates_lag` must be smaller than/equal to `max_covariates_lag`."
                    ),
                    logger=logger,
                )


class PastCovariatesIndexGenerator(CovariatesIndexGenerator):
    """Generates index for past covariates on train and inference datasets"""

    def generate_train_idx(
        self, target: TimeSeries, covariates: Optional[TimeSeries] = None
    ) -> Tuple[SupportedIndex, pd.Timestamp]:

        super().generate_train_idx(target, covariates)

        # the returned index depends on the following cases:
        # case 0
        #     user supplied covariates: simply return the covariate time index; guarantees that an exception is
        #     raised if user supplied insufficient covariates
        # case 1
        #     only input_chunk_length and output_chunk_length are given: the complete covariate index is within the
        #     target index; always True for all models except RegressionModels.
        # case 2
        #     covariate lags were given (shift_start <= 0 and shift_end <= 0) and
        #     abs(shift_start - 1) <= input_chunk_length: the complete covariate index is within the target index;
        #     can only be True for RegressionModels.
        # case 3
        #     covariate lags were given (shift_start <= 0 and shift_end <= 0) and
        #     abs(shift_start - 1) > input_chunk_length: we need to add indices before the beginning of the target
        #     series; can only be True for RegressionModels.

        target_end = target.end_time()
        if covariates is not None:  # case 0
            return covariates.time_index, target_end

        if self.shift_start is None:  # case 1
            steps_ahead_start = 0
        else:  # case 2 & 3
            steps_ahead_start = self.input_chunk_length + (self.shift_start - 1)

        if not self.shift_end:  # case 1
            steps_ahead_end = -self.output_chunk_length
        else:  # case 2 & 3
            steps_ahead_end = -(self.output_chunk_length - self.shift_end)

        steps_ahead_end = steps_ahead_end if steps_ahead_end else None
        return (
            _generate_train_idx(target, steps_ahead_start, steps_ahead_end),
            target_end,
        )

    def generate_inference_idx(
        self, n: int, target: TimeSeries, covariates: Optional[TimeSeries] = None
    ) -> Tuple[SupportedIndex, pd.Timestamp]:

        super().generate_inference_idx(n, target, covariates)

        # for prediction (`n` is given) with past covariates the returned index depends on the following cases:
        # case 0
        #     user supplied covariates: simply return the covariate time index; guarantees that an exception is
        #     raised if user supplied insufficient covariates.
        # case 1
        #     only input_chunk_length and output_chunk_length are given: we need to generate a time index that starts
        #     `input_chunk_length - 1` before the end of `target` and ends `max(0, n - output_chunk_length)` after the
        #     end of `target`; always True for all models except RegressionModels.
        # case 2
        #     covariate lags were given (shift_start <= 0 and shift_end <= 0): we need to generate a time index that
        #     starts `-shift_start` before the end of `target` and has a length of
        #     `shift_steps + max(0, n - output_chunk_length)`, where `shift_steps` is the number of time steps between
        #     `shift_start` and `shift_end`; can only be True for RegressionModels.

        target_end = target.end_time()
        if covariates is not None:  # case 0
            return covariates.time_index, target_end

        if self.shift_start is None or self.shift_end is None:  # case 1
            steps_back_end = self.input_chunk_length - 1
            n_steps = steps_back_end + 1 + max(0, n - self.output_chunk_length)
        else:  # case 2
            steps_back_end = -self.shift_start
            shift_steps = self.shift_end - self.shift_start + 1
            n_steps = shift_steps + max(0, n - self.output_chunk_length)

        return (
            generate_index(
                start=target.end_time() - target.freq * steps_back_end,
                length=n_steps,
                freq=target.freq,
            ),
            target_end,
        )

    @property
    def base_component_name(self) -> str:
        return "pc"

    def _verify_lags(self, min_covariates_lag, max_covariates_lag):
        # general lag checks
        super()._verify_lags(min_covariates_lag, max_covariates_lag)
        # check past covariate specific lag requirements
        if min_covariates_lag is not None and min_covariates_lag >= 0:
            raise_log(ValueError("`min_covariates_lag` must be < 0."), logger=logger)

        if max_covariates_lag is not None and max_covariates_lag >= 0:
            raise_log(ValueError("`max_covariates_lag` must be < 0."), logger=logger)


class FutureCovariatesIndexGenerator(CovariatesIndexGenerator):
    """Generates index for future covariates on train and inference datasets."""

    def generate_train_idx(
        self, target: TimeSeries, covariates: Optional[TimeSeries] = None
    ) -> Tuple[SupportedIndex, pd.Timestamp]:

        super().generate_train_idx(target, covariates)

        # the returned index depends on the following cases:
        # case 0
        #     user supplied covariates: simply return the covariate time index; guarantees that models raise an
        #     exception if user supplied insufficient covariates
        # case 1
        #     user uses a LocalForecastingModel or model agnostic scenario (input_chunk_length is None):
        #     simply return the target time index.
        # case 2
        #     only input_chunk_length and output_chunk_length are given: the complete covariate index is within the
        #     target index; always True for all models except RegressionModels.
        # case 3
        #     covariate lags were given and (shift_start <= 0 or shift_end <= 0): historic part of future covariates.
        #     if shift_end < 0 there will only be the historic part of future covariates.
        #     If shift_start <= 0 and abs(shift_start - 1) > input_chunk_length: we need to add indices before the
        #     beginning of the target series; can only be True for RegressionModels.
        # case 4
        #     covariate lags were given and (shift_start > 0 or shift_end > 0): future part of future covariates.
        #     if shift_start > 0 there will only be the future part of future covariates.
        #     If shift_end > 0 and shift_start > input_chunk_length: we need to add indices after the end of the
        #     target series; can only be True for RegressionModels.

        target_end = target.end_time()

        if covariates is not None:  # case 0
            return covariates.time_index, target_end

        if self.input_chunk_length is None:  # case 1
            return target.time_index, target_end

        if self.shift_start is None:  # case 2
            steps_ahead_start = 0
        else:  # case 3
            steps_ahead_start = self.input_chunk_length + self.shift_start - 1

        if self.shift_end is None:  # case 2
            steps_ahead_end = 0
        else:  # case 4
            steps_ahead_end = -self.output_chunk_length + self.shift_end
        steps_ahead_end = steps_ahead_end if steps_ahead_end else None

        return (
            _generate_train_idx(target, steps_ahead_start, steps_ahead_end),
            target_end,
        )

    def generate_inference_idx(
        self, n: int, target: TimeSeries, covariates: Optional[TimeSeries] = None
    ) -> Tuple[SupportedIndex, pd.Timestamp]:

        super().generate_inference_idx(n, target, covariates)

        # for prediction (`n` is given) with future covariates the returned index depends on the following cases:
        # case 0
        #     user supplied covariates: simply return the covariate time index; guarantees that an exception is
        #     raised if user supplied insufficient covariates
        # case 1
        #     user uses a LocalForecastingModel or model agnostic scenario (input_chunk_length is None):
        #     simply return the target time index.
        # case 2
        #     only input_chunk_length and output_chunk_length are given: we need to generate a time index that starts
        #     `input_chunk_length - 1` before the end of `target` and ends `max(n, output_chunk_length)` after the
        #     end of `target`; always True for all models except RegressionModels.
        # case 3
        #     covariate lags were given: we need to generate a time index that starts `-shift_start`
        #     steps before the end of `target` and has a length of `shift_steps + max(0, n - output_chunk_length)`,
        #     where `shift_steps` is `shift_end - shift_start`; can only be True for RegressionModels.

        target_end = target.end_time()
        if covariates is not None:  # case 0
            return covariates.time_index, target_end

        if self.input_chunk_length is None:
            steps_back_end = -1
            n_steps = n
        elif self.shift_start is None:  # case 2
            steps_back_end = self.input_chunk_length - 1
            n_steps = steps_back_end + 1 + max(n, self.output_chunk_length)
        else:  # case 3
            steps_back_end = -self.shift_start
            shift_steps = self.shift_end + steps_back_end + 1
            n_steps = shift_steps + max(0, n - self.output_chunk_length)

        return (
            generate_index(
                start=target.end_time() - target.freq * steps_back_end,
                length=n_steps,
                freq=target.freq,
            ),
            target_end,
        )

    @property
    def base_component_name(self) -> str:
        return "fc"


class Encoder(ABC):
    """Abstract class for all encoders"""

    @abstractmethod
    def __init__(self):
        self.attribute = None
        self.dtype = np.float64
        self._fit_called = False

    @abstractmethod
    def encode_train(
        self,
        target: TimeSeries,
        covariates: Optional[TimeSeries] = None,
        merge_covariates: bool = True,
        **kwargs,
    ) -> TimeSeries:
        """Each subclass must implement a method to encode the covariates index for training.

        Parameters
        ----------
        target
            The target TimeSeries used during training or passed to prediction as `series`.
        covariates
            Optionally, the past or future covariates used for training.
        merge_covariates
            Whether or not to merge the encoded TimeSeries with `covariates`.
        """
        pass

    @abstractmethod
    def encode_inference(
        self,
        n: int,
        target: TimeSeries,
        covariates: Optional[TimeSeries] = None,
        merge_covariates: bool = True,
        **kwargs,
    ) -> TimeSeries:
        """Each subclass must implement a method to encode the covariates index for prediction.

        Parameters
        ----------
        n
            The forecast horizon
        target
            The target TimeSeries used during training or passed to prediction as `series`
        covariates
            Optionally, the past or future covariates used for prediction.
        merge_covariates
            Whether or not to merge the encoded TimeSeries with `covariates`.

        """
        pass

    @staticmethod
    def _merge_covariates(
        encoded: TimeSeries, covariates: Optional[TimeSeries] = None
    ) -> TimeSeries:
        """If (actual) covariates are given, merge the encoded index with the covariates

        Parameters
        ----------
        encoded
            The encoded TimeSeries either from `encode_train()` or `encode_inference()`
        covariates
            Optionally, some past or future covariates supplied by the user.
        """
        return covariates.stack(encoded) if covariates is not None else encoded

    @staticmethod
    def _drop_encoded_components(
        covariates: Optional[TimeSeries], components: pd.Index
    ) -> Optional[TimeSeries]:
        """Avoid pitfalls: `encode_train()` or `encode_inference()` can be called multiple times or chained.
        Exclude any encoded components from `covariates` to generate and add the new encodings at a later time.
        """
        if covariates is None:
            return covariates

        duplicate_components = components[components.isin(covariates.components)]
        # case 1: covariates only consist of encoded components
        if len(duplicate_components) == len(covariates.components):
            covariates = None
        # case 2: covariates also have non-encoded components
        elif len(duplicate_components) and len(duplicate_components) < len(
            covariates.components
        ):
            covariates = covariates[
                list(
                    covariates.components[
                        ~covariates.components.isin(duplicate_components)
                    ]
                )
            ]
        return covariates

    @property
    def fit_called(self) -> bool:
        """Returns whether the `Encoder` object has been fitted."""
        return self._fit_called

    @property
    @abstractmethod
    def requires_fit(self) -> bool:
        """Whether the `Encoder` sub class must be fit with `Encoder.encode_train()` before inference
        with `Encoder.encode_inference()`."""
        pass


class SingleEncoder(Encoder, ABC):
    """`SingleEncoder`: Abstract class for single index encoders.
    Single encoders can be used to implement new encoding techniques.
    Each single encoder must implement an `_encode()` method that carries the encoding logic.

    The `_encode()` method must take an `index` as input and generate a encoded single `TimeSeries` as output.
    """

    def __init__(self, index_generator: CovariatesIndexGenerator):
        """Single encoders take an `index_generator` to generate the required index for encoding past and future
        covariates.
        See darts.utils.data.covariate_index_generators.py for the `CovariatesIndexGenerator` subclasses.
        For past covariates encoders, use a `PastCovariatesIndexGenerator`.
        For future covariates encoders use a `FutureCovariatesIndexGenerator`.

        Parameters
        ----------
        index_generator
            An instance of `CovariatesIndexGenerator` with methods `generate_train_idx()` and
            `generate_inference_idx()`. Used to generate the index for encoders.
        """

        super().__init__()
        self.index_generator = index_generator
        self._components = pd.Index([])

    @abstractmethod
    def _encode(
        self, index: SupportedIndex, target_end: pd.Timestamp, dtype: np.dtype
    ) -> TimeSeries:
        """Single Encoders must implement an _encode() method to encode the index.

        Parameters
        ----------
        index
            The index generated from `self.index_generator` for either the train or inference dataset.
        target_end
            The end time of the target series.
        dtype
            The dtype of the encoded index
        """
        pass

    def encode_train(
        self,
        target: TimeSeries,
        covariates: Optional[TimeSeries] = None,
        merge_covariates: bool = True,
        **kwargs,
    ) -> TimeSeries:
        """Returns encoded index for training.

        Parameters
        ----------
        target
            The target TimeSeries used during training or passed to prediction as `series`
        covariates
            Optionally, the covariates used for training: past covariates if `self.index_generator` is a
            `PastCovariatesIndexGenerator`, future covariates if `self.index_generator` is a
            `FutureCovariatesIndexGenerator`
        merge_covariates
            Whether or not to merge the encoded TimeSeries with `covariates`.
        """
        # exclude encoded components from covariates to add the newly encoded components later
        covariates = self._drop_encoded_components(covariates, self.components)

        # generate index and encodings
        index, target_end = self.index_generator.generate_train_idx(target, covariates)
        encoded = self._encode(index, target_end, target.dtype)

        # optionally, merge encodings with original `covariates` series
        encoded = (
            self._merge_covariates(encoded, covariates=covariates)
            if merge_covariates
            else encoded
        )

        # save encoded component names
        if self.components.empty:
            components = encoded.components
            if covariates is not None:
                components = components[~components.isin(covariates.components)]
            self._components = components

        self._fit_called = True
        return encoded

    def encode_inference(
        self,
        n: int,
        target: TimeSeries,
        covariates: Optional[TimeSeries] = None,
        merge_covariates: bool = True,
        **kwargs,
    ) -> TimeSeries:
        """Returns encoded index for inference/prediction.

        Parameters
        ----------
        n
            The forecast horizon
        target
            The target TimeSeries used during training or passed to prediction as `series`
        covariates
            Optionally, the covariates used for prediction: past covariates if `self.index_generator` is a
            `PastCovariatesIndexGenerator`, future covariates if `self.index_generator` is a
            `FutureCovariatesIndexGenerator`
        merge_covariates
            Whether or not to merge the encoded TimeSeries with `covariates`.
        """
        # some encoders must be fit before `encode_inference()`
        raise_if(
            not self.fit_called and self.requires_fit,
            f"`{self.__class__.__name__}` object must be trained before inference. "
            f"Call method `encode_train()` before `encode_inference()`.",
            logger=logger,
        )

        # exclude encoded components from covariates to add the newly encoded components later
        covariates = self._drop_encoded_components(covariates, self.components)

        # generate index and encodings
        index, target_end = self.index_generator.generate_inference_idx(
            n, target, covariates
        )
        encoded = self._encode(index, target_end, target.dtype)

        # optionally, merge encodings with original `covariates` series
        encoded = (
            self._merge_covariates(encoded, covariates=covariates)
            if merge_covariates
            else encoded
        )

        # optionally, save encoded component names also at inference as some encoders do not have to be trained before
        if self.components.empty:
            components = encoded.components
            if covariates is not None:
                components = components[~components.isin(covariates.components)]
            self._components = components

        return encoded

    @property
    @abstractmethod
    def accept_transformer(self) -> List[bool]:
        """Whether the `SingleEncoder` sub class accepts to be transformed."""
        pass

    @property
    def components(self) -> pd.Index:
        """Returns the encoded component names. Only available after `Encoder.encode_train()` or
        `Encoder.encode_inference()` have been called."""
        return self._components

    @property
    @abstractmethod
    def base_component_name(self) -> str:
        """Returns the base encoder base component name. The string follows the given format:
        `"darts_enc_{covariates_temp}_{encoder}_{attribute}"`, where the elements are:

        * covariates_temp: "pc" or "fc" for past, or future covariates respectively.
        * encoder: the SingleEncoder type used:
            * "cyc" (cyclic temporal encoder),
            * "dta" (datetime attribute encoder),
            * "pos" (positional integer index encoder),
            * "cus" (custom callable index encoder)
        * attribute: the attribute used for the underlying encoder. Some examples:
            * "month_sin", "month_cos" (for "cyc")
            * "month" (for "dta")
            * "relative" (for "pos")
            * "custom" (for "cus")
        """
        return f"darts_enc_{self.index_generator.base_component_name}"


class SequentialEncoderTransformer:
    """`SequentialEncoderTransformer` applies transformation to the non-transformed encoded covariates output of
    `SequentialEncoder.encode_train()` and `SequentialEncoder.encode_inference()`. The transformer is fitted
    when `transform()` is called for the first time. This ensures proper transformation of train, validation and
    inference dataset covariates. User-supplied covariates are not transformed."""

    def __init__(
        self, transformer: FittableDataTransformer, transform_mask: List[bool]
    ):
        """
        Parameters
        ----------
        transformer
            A `FittableDataTransformer` object with a `fit_transform()` and `transform()` method.
        transform_mask
            A boolean 1-D mask specifying which of the input covariates to :meth:`transform()
            <SequentialEncoderTransformer.transform()>` must be transformed.
        """
        self.transformer: FittableDataTransformer = transformer
        self.transform_mask: np.ndarray = np.array(transform_mask)
        self._fit_called: bool = False

    def transform(self, covariates: List[TimeSeries]) -> List[TimeSeries]:
        """This method applies transformation to the non-transformed encoded covariates output of
        `SequentialEncoder._encode_sequence()` after being merged with user-defined covariates. The transformer is
        fitted when `transform()` is called for the first time. This ensures proper transformation of train, validation
        and inference dataset covariates. The masks ensure that no covariates are transformed that user explicitly
        supplied to `TorchForecastingModel.fit()` and `TorchForecastingModel.predict()`

        Parameters
        ----------
        covariates
            The non-transformed encoded covariates output of `SequentialEncoder._encode_sequence()` before merging with
            user-defined covariates.
        """
        if not self.fit_called:
            self._update_mask(covariates)
            if any(self.transform_mask):
                # fit the transformer on all encoded values by concatenating multi-series input encodings
                self.transformer.fit(
                    series=TimeSeries.from_values(
                        np.concatenate([cov.values() for cov in covariates]),
                        columns=covariates[0].components,
                    ),
                    component_mask=self.transform_mask,
                )
            self._fit_called = True

        if any(self.transform_mask):
            transformed = [
                self.transformer.transform(cov, component_mask=self.transform_mask)
                for cov in covariates
            ]
        else:
            transformed = covariates
        return transformed

    def _update_mask(self, covariates: List[TimeSeries]) -> None:
        """if user supplied additional covariates to model.fit() or model.predict(), `self.transform_mask` has to be
        updated as user-defined covariates should not be transformed. These covariates are always located in the
        first `n_diff = covariates[0].width - len(self.transform_mask)` components of each TimeSeries in in
        `covariates`.
        """

        n_diff = covariates[0].width - len(self.transform_mask)
        if not n_diff:
            pass
        else:
            self.transform_mask = np.array([False] * n_diff + list(self.transform_mask))

    @property
    def fit_called(self) -> bool:
        """Return whether or not the transformer has been fitted."""
        return self._fit_called


def _generate_train_idx(target, steps_ahead_start, steps_ahead_end) -> SupportedIndex:
    """The returned index depends on the following cases:

    case 1
        (steps_ahead_start >= 0 and steps_ahead_end is None or <= 1)
        the complete index is within the target index; always True for all models except RegressionModels.
    case 2
        steps_ahead_start < 0: add indices before the target start time; only possible for RegressionModels
        where the minimum past lag is larger than input_chunk_length.
    case 3
        steps_ahead_end > 0: add indices after the target end time; only possible for RegressionModels
        where the maximum future lag is larger than output_chunk_length.

    Parameters
    ----------
    target
        the target series.
    steps_ahead_start
        how many steps ahead of target start time to begin the index.
    steps_ahead_end
        how many steps ahead of target end time to end the index.
    """
    # case 1
    if steps_ahead_start >= 0 and (steps_ahead_end is None or steps_ahead_end <= -1):
        return target.time_index[steps_ahead_start:steps_ahead_end]

    # case 2
    idx_start = (
        generate_index(
            end=target.start_time() - target.freq,
            length=abs(steps_ahead_start),
            freq=target.freq,
        )
        if steps_ahead_start < 0
        else target.time_index.__class__([])
    )

    # if `steps_ahead_start >= 0` or `steps_ahead_end <= 0` we must extract a slice of the target series index
    center_start = steps_ahead_start if steps_ahead_start >= 0 else None
    center_end = (
        steps_ahead_end
        if steps_ahead_end is not None and steps_ahead_end <= 0
        else None
    )
    idx_center = target.time_index[center_start:center_end]

    # case 3
    idx_end = (
        generate_index(
            start=target.end_time() + target.freq,
            length=abs(steps_ahead_end),
            freq=target.freq,
        )
        if steps_ahead_end is not None and steps_ahead_end > 0
        else target.time_index.__class__([])
    )

    # concatenate start, center, and end index
    # note: pandas' union() returns type pd.Index(), so we construct index directly from index class
    return target.time_index.__class__(idx_start.union(idx_center).union(idx_end))



"""
TimeSeries Datasets
-------------------
"""

try:
    # Base classes for training datasets:
    # Implementation (horizon-based)
    from .horizon_based_dataset import HorizonBasedDataset

    # Base class and implementations for inference datasets:
    from .inference_dataset import (
        DualCovariatesInferenceDataset,
        FutureCovariatesInferenceDataset,
        InferenceDataset,
        MixedCovariatesInferenceDataset,
        PastCovariatesInferenceDataset,
        SplitCovariatesInferenceDataset,
    )

    # Implementations (sequential)
    from .sequential_dataset import (
        DualCovariatesSequentialDataset,
        FutureCovariatesSequentialDataset,
        MixedCovariatesSequentialDataset,
        PastCovariatesSequentialDataset,
        SplitCovariatesSequentialDataset,
    )

    # Implementations (shifted)
    from .shifted_dataset import (
        DualCovariatesShiftedDataset,
        FutureCovariatesShiftedDataset,
        MixedCovariatesShiftedDataset,
        PastCovariatesShiftedDataset,
        SplitCovariatesShiftedDataset,
    )
    from .training_dataset import (
        DualCovariatesTrainingDataset,
        FutureCovariatesTrainingDataset,
        MixedCovariatesTrainingDataset,
        PastCovariatesTrainingDataset,
        SplitCovariatesTrainingDataset,
        TrainingDataset,
    )

except ImportError:
    # Torch is not available
    pass



import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.logging import get_logger
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.timeseries_generation import gaussian_timeseries

logger = get_logger(__name__)

try:
    from darts.utils.data import (  # noqa: F401
        DualCovariatesInferenceDataset,
        DualCovariatesSequentialDataset,
        DualCovariatesShiftedDataset,
        FutureCovariatesInferenceDataset,
        FutureCovariatesSequentialDataset,
        FutureCovariatesShiftedDataset,
        HorizonBasedDataset,
        MixedCovariatesInferenceDataset,
        MixedCovariatesSequentialDataset,
        MixedCovariatesShiftedDataset,
        PastCovariatesInferenceDataset,
        PastCovariatesSequentialDataset,
        PastCovariatesShiftedDataset,
        SplitCovariatesInferenceDataset,
        SplitCovariatesSequentialDataset,
        SplitCovariatesShiftedDataset,
    )

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not installed - dataset tests will be skipped.")
    TORCH_AVAILABLE = False

if TORCH_AVAILABLE:

    class DatasetTestCase(DartsBaseTestClass):
        target1 = gaussian_timeseries(length=100).with_static_covariates(
            pd.Series([0, 1], index=["st1", "st2"])
        )
        target2 = gaussian_timeseries(length=150).with_static_covariates(
            pd.Series([2, 3], index=["st1", "st2"])
        )
        cov_st1 = target1.static_covariates.values
        cov_st2 = target2.static_covariates.values
        cov_st2_df = pd.Series([2, 3], index=["st1", "st2"])
        vals1, vals2 = target1.values(), target2.values()
        cov1, cov2 = gaussian_timeseries(length=100), gaussian_timeseries(length=150)

        def _assert_eq(self, lefts: tuple, rights: tuple):
            for left, right in zip(lefts, rights):
                left = left.values() if isinstance(left, TimeSeries) else left
                right = right.values() if isinstance(right, TimeSeries) else right
                assert type(left) == type(right)
                assert (
                    isinstance(
                        left, (TimeSeries, pd.Series, pd.DataFrame, np.ndarray, list)
                    )
                    or left is None
                )
                if isinstance(left, (pd.Series, pd.DataFrame)):
                    assert left.equals(right)
                elif isinstance(left, np.ndarray):
                    assert np.array_equal(left, right)
                elif isinstance(left, (list, TimeSeries)):
                    assert left == right
                else:
                    assert right is None

        def test_past_covariates_inference_dataset(self):
            # one target series
            ds = PastCovariatesInferenceDataset(
                target_series=self.target1, input_chunk_length=len(self.target1)
            )
            np.testing.assert_almost_equal(ds[0][0], self.vals1)
            self._assert_eq(ds[0][1:], (None, None, self.cov_st1, self.target1))

            # two target series
            ds = PastCovariatesInferenceDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=max(len(self.target1), len(self.target2)),
            )
            np.testing.assert_almost_equal(ds[1][0], self.vals2)
            self._assert_eq(ds[1][1:], (None, None, self.cov_st2, self.target2))

            # fail if covariates do not have same size
            with self.assertRaises(ValueError):
                ds = PastCovariatesInferenceDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # with covariates
            ds = PastCovariatesInferenceDataset(
                target_series=[self.target1, self.target2],
                covariates=[self.cov1, self.cov2],
                input_chunk_length=max(len(self.target1), len(self.target2)),
            )
            np.testing.assert_almost_equal(ds[1][0], self.vals2)
            np.testing.assert_almost_equal(ds[1][1], self.cov2.values())
            self._assert_eq(
                ds[1][2:], (None, self.cov_st2, self.target2)
            )  # no "future past" covariate here

            # more complex case with future past covariates:
            times1 = pd.date_range(start="20100101", end="20100701", freq="D")
            times2 = pd.date_range(
                start="20100101", end="20100820", freq="D"
            )  # 50 days longer than times1

            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            short_cov = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            )
            long_cov = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )

            ds = PastCovariatesInferenceDataset(
                target_series=target,
                covariates=short_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            # should fail if covariates are too short
            with self.assertRaises(ValueError):
                _ = ds[0]

            # Should return correct values when covariates is long enough
            ds = PastCovariatesInferenceDataset(
                target_series=target,
                covariates=long_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-60:-50])
            np.testing.assert_almost_equal(ds[0][2], long_cov.values()[-50:-30])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            self.assertEqual(ds[0][4], target)

            # Should also work for integer-indexed series
            target = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
            ).with_static_covariates(self.cov_st2_df)
            covariate = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
            )

            ds = PastCovariatesInferenceDataset(
                target_series=target,
                covariates=covariate,
                input_chunk_length=10,
                output_chunk_length=10,
                n=20,
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], covariate.values()[20:30])
            np.testing.assert_almost_equal(ds[0][2], covariate.values()[30:40])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            self.assertEqual(ds[0][4], target)

        def test_future_covariates_inference_dataset(self):
            # one target series
            ds = FutureCovariatesInferenceDataset(
                target_series=self.target1, input_chunk_length=len(self.target1)
            )
            np.testing.assert_almost_equal(ds[0][0], self.vals1)
            self._assert_eq(ds[0][1:], (None, self.cov_st1, self.target1))

            # two target series
            ds = FutureCovariatesInferenceDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=max(len(self.target1), len(self.target2)),
            )
            np.testing.assert_almost_equal(ds[1][0], self.vals2)
            self._assert_eq(ds[1][1:], (None, self.cov_st2, self.target2))

            # fail if covariates do not have same size
            with self.assertRaises(ValueError):
                ds = FutureCovariatesInferenceDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # With future past covariates:
            times1 = pd.date_range(start="20100101", end="20100701", freq="D")
            times2 = pd.date_range(
                start="20100101", end="20100820", freq="D"
            )  # 50 days longer than times1

            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            short_cov = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            )
            long_cov = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )

            ds = FutureCovariatesInferenceDataset(
                target_series=target, covariates=short_cov, input_chunk_length=10, n=30
            )

            # should fail if covariates are too short
            with self.assertRaises(ValueError):
                _ = ds[0]

            # Should return correct values when covariates is long enough
            ds = FutureCovariatesInferenceDataset(
                target_series=target, covariates=long_cov, input_chunk_length=10, n=30
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-50:-20])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            self.assertEqual(ds[0][3], target)

            # Should also work for integer-indexed series
            target = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
            ).with_static_covariates(self.cov_st2_df)
            covariate = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
            )

            ds = FutureCovariatesInferenceDataset(
                target_series=target, covariates=covariate, input_chunk_length=10, n=20
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], covariate.values()[30:50])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            self.assertEqual(ds[0][3], target)

        def test_dual_covariates_inference_dataset(self):
            # one target series
            ds = DualCovariatesInferenceDataset(
                target_series=self.target1, input_chunk_length=len(self.target1)
            )
            np.testing.assert_almost_equal(ds[0][0], self.vals1)
            self._assert_eq(ds[0][1:], (None, None, self.cov_st1, self.target1))

            # two target series
            ds = DualCovariatesInferenceDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=max(len(self.target1), len(self.target2)),
            )
            np.testing.assert_almost_equal(ds[1][0], self.vals2)
            self._assert_eq(ds[1][1:], (None, None, self.cov_st2, self.target2))

            # fail if covariates do not have same size
            with self.assertRaises(ValueError):
                ds = DualCovariatesInferenceDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # With future past covariates:
            times1 = pd.date_range(start="20100101", end="20100701", freq="D")
            times2 = pd.date_range(
                start="20100101", end="20100820", freq="D"
            )  # 50 days longer than times1

            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            short_cov = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            )
            long_cov = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )

            ds = DualCovariatesInferenceDataset(
                target_series=target,
                covariates=short_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            # should fail if covariates are too short
            with self.assertRaises(ValueError):
                _ = ds[0]

            # Should return correct values when covariates is long enough
            ds = DualCovariatesInferenceDataset(
                target_series=target,
                covariates=long_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-60:-50])
            np.testing.assert_almost_equal(ds[0][2], long_cov.values()[-50:-20])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            self.assertEqual(ds[0][4], target)

            # Should also work for integer-indexed series
            target = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
            ).with_static_covariates(self.cov_st2_df)
            covariate = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
            )

            ds = DualCovariatesInferenceDataset(
                target_series=target,
                covariates=covariate,
                input_chunk_length=10,
                output_chunk_length=10,
                n=20,
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], covariate.values()[20:30])
            np.testing.assert_almost_equal(ds[0][2], covariate.values()[30:50])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            self.assertEqual(ds[0][4], target)

        def test_mixed_covariates_inference_dataset(self):
            # With future past covariates:
            times1 = pd.date_range(start="20100101", end="20100701", freq="D")
            times2 = pd.date_range(
                start="20100201", end="20100820", freq="D"
            )  # ends 50 days after times1

            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            past_cov = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            )
            long_past_cov = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )
            future_cov = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )

            ds = MixedCovariatesInferenceDataset(
                target_series=target,
                past_covariates=past_cov,
                future_covariates=past_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            # should fail if future covariates are too short
            with self.assertRaises(ValueError):
                _ = ds[0]

            # Should return correct values when covariates is long enough
            ds = MixedCovariatesInferenceDataset(
                target_series=target,
                past_covariates=long_past_cov,
                future_covariates=future_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            # It should contain:
            # past_target, past_covariates, historic_future_covariates, future_covariates, future_past_covariates
            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
            np.testing.assert_almost_equal(ds[0][2], future_cov.values()[-60:-50])
            np.testing.assert_almost_equal(ds[0][3], future_cov.values()[-50:-20])
            np.testing.assert_almost_equal(ds[0][4], long_past_cov.values()[-50:-30])
            np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
            self.assertEqual(ds[0][6], target)

            # Should also work for integer-indexed series
            target = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
            ).with_static_covariates(self.cov_st2_df)
            past_cov = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
            )
            future_cov = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70)
            )

            ds = MixedCovariatesInferenceDataset(
                target_series=target,
                past_covariates=past_cov,
                future_covariates=future_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=20,
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
            np.testing.assert_almost_equal(ds[0][2], future_cov.values()[10:20])
            np.testing.assert_almost_equal(ds[0][3], future_cov.values()[20:40])
            np.testing.assert_almost_equal(ds[0][4], past_cov.values()[30:40])
            np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
            self.assertEqual(ds[0][6], target)

        def test_split_covariates_inference_dataset(self):
            # With future past covariates:
            times1 = pd.date_range(start="20100101", end="20100701", freq="D")
            times2 = pd.date_range(
                start="20100201", end="20100820", freq="D"
            )  # ends 50 days after times1

            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            past_cov = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            )
            long_past_cov = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )
            future_cov = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )

            ds = SplitCovariatesInferenceDataset(
                target_series=target,
                past_covariates=past_cov,
                future_covariates=past_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            # should fail if future covariates are too short
            with self.assertRaises(ValueError):
                _ = ds[0]

            # Should return correct values when covariates is long enough
            ds = SplitCovariatesInferenceDataset(
                target_series=target,
                past_covariates=long_past_cov,
                future_covariates=future_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=30,
            )

            # It should contain:
            # past_target, past_covariates, future_covariates, future_past_covariates
            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
            np.testing.assert_almost_equal(ds[0][2], future_cov.values()[-50:-20])
            np.testing.assert_almost_equal(ds[0][3], long_past_cov.values()[-50:-30])
            np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
            self.assertEqual(ds[0][5], target)

            # Should also work for integer-indexed series
            target = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
            ).with_static_covariates(self.cov_st2_df)
            past_cov = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
            )
            future_cov = TimeSeries.from_times_and_values(
                pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70)
            )

            ds = SplitCovariatesInferenceDataset(
                target_series=target,
                past_covariates=past_cov,
                future_covariates=future_cov,
                input_chunk_length=10,
                output_chunk_length=10,
                n=20,
            )

            np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
            np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
            np.testing.assert_almost_equal(ds[0][2], future_cov.values()[20:40])
            np.testing.assert_almost_equal(ds[0][3], past_cov.values()[30:40])
            np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
            self.assertEqual(ds[0][5], target)

        def test_past_covariates_sequential_dataset(self):
            # one target series
            ds = PastCovariatesSequentialDataset(
                target_series=self.target1,
                input_chunk_length=10,
                output_chunk_length=10,
            )
            self.assertEqual(len(ds), 81)
            self._assert_eq(
                ds[5], (self.target1[75:85], None, self.cov_st1, self.target1[85:95])
            )

            # two target series
            ds = PastCovariatesSequentialDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=10,
                output_chunk_length=10,
            )
            self.assertEqual(len(ds), 262)
            self._assert_eq(
                ds[5], (self.target1[75:85], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[136],
                (self.target2[125:135], None, self.cov_st2, self.target2[135:145]),
            )

            # two target series with custom max_nr_samples
            ds = PastCovariatesSequentialDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=10,
                output_chunk_length=10,
                max_samples_per_ts=50,
            )
            self.assertEqual(len(ds), 100)
            self._assert_eq(
                ds[5], (self.target1[75:85], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[55],
                (self.target2[125:135], None, self.cov_st2, self.target2[135:145]),
            )

            # two targets and one covariate
            with self.assertRaises(ValueError):
                ds = PastCovariatesSequentialDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # two targets and two covariates
            ds = PastCovariatesSequentialDataset(
                target_series=[self.target1, self.target2],
                covariates=[self.cov1, self.cov2],
                input_chunk_length=10,
                output_chunk_length=10,
            )
            self._assert_eq(
                ds[5],
                (
                    self.target1[75:85],
                    self.cov1[75:85],
                    self.cov_st1,
                    self.target1[85:95],
                ),
            )
            self._assert_eq(
                ds[136],
                (
                    self.target2[125:135],
                    self.cov2[125:135],
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # should fail if covariates do not have the required time span, even though covariates are longer
            times1 = pd.date_range(start="20100101", end="20110101", freq="D")
            times2 = pd.date_range(start="20120101", end="20150101", freq="D")
            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
            ds = PastCovariatesSequentialDataset(
                target_series=target,
                covariates=cov,
                input_chunk_length=10,
                output_chunk_length=10,
            )
            with self.assertRaises(ValueError):
                _ = ds[5]

            # the same should fail when series are integer-indexed
            times1 = pd.RangeIndex(start=0, stop=100, step=1)
            times2 = pd.RangeIndex(start=200, stop=400, step=1)
            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
            ds = PastCovariatesSequentialDataset(
                target_series=target,
                covariates=cov,
                input_chunk_length=10,
                output_chunk_length=10,
            )
            with self.assertRaises(ValueError):
                _ = ds[5]

            # we should get the correct covariate slice even when target and covariates are not aligned
            times1 = pd.date_range(start="20100101", end="20110101", freq="D")
            times2 = pd.date_range(start="20090101", end="20110106", freq="D")
            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
            ds = PastCovariatesSequentialDataset(
                target_series=target,
                covariates=cov,
                input_chunk_length=10,
                output_chunk_length=10,
            )

            np.testing.assert_almost_equal(ds[0][1], cov.values()[-25:-15])
            np.testing.assert_almost_equal(ds[5][1], cov.values()[-30:-20])

            # This should also be the case when series are integer indexed
            times1 = pd.RangeIndex(start=100, stop=200, step=1)
            times2 = pd.RangeIndex(start=50, stop=250, step=1)
            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
            ds = PastCovariatesSequentialDataset(
                target_series=target,
                covariates=cov,
                input_chunk_length=10,
                output_chunk_length=10,
            )

            np.testing.assert_almost_equal(ds[0][1], cov.values()[-70:-60])
            np.testing.assert_almost_equal(ds[5][1], cov.values()[-75:-65])

        def test_future_covariates_sequential_dataset(self):
            # one target series
            ds = FutureCovariatesSequentialDataset(
                target_series=self.target1,
                input_chunk_length=10,
                output_chunk_length=10,
            )
            self.assertEqual(len(ds), 81)
            self._assert_eq(
                ds[5], (self.target1[75:85], None, self.cov_st1, self.target1[85:95])
            )

            # two target series
            ds = FutureCovariatesSequentialDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=10,
                output_chunk_length=10,
            )
            self.assertEqual(len(ds), 262)
            self._assert_eq(
                ds[5], (self.target1[75:85], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[136],
                (self.target2[125:135], None, self.cov_st2, self.target2[135:145]),
            )

            # two target series with custom max_nr_samples
            ds = FutureCovariatesSequentialDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=10,
                output_chunk_length=10,
                max_samples_per_ts=50,
            )
            self.assertEqual(len(ds), 100)
            self._assert_eq(
                ds[5], (self.target1[75:85], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[55],
                (self.target2[125:135], None, self.cov_st2, self.target2[135:145]),
            )

            # two targets and one covariate
            with self.assertRaises(ValueError):
                ds = FutureCovariatesSequentialDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # two targets and two covariates; covariates not aligned, must contain correct values
            target1 = TimeSeries.from_values(
                np.random.randn(100)
            ).with_static_covariates(self.cov_st2_df)
            target2 = TimeSeries.from_values(
                np.random.randn(50)
            ).with_static_covariates(self.cov_st2_df)
            cov1 = TimeSeries.from_values(np.random.randn(120))
            cov2 = TimeSeries.from_values(np.random.randn(80))

            ds = FutureCovariatesSequentialDataset(
                target_series=[target1, target2],
                covariates=[cov1, cov2],
                input_chunk_length=10,
                output_chunk_length=10,
            )

            np.testing.assert_almost_equal(ds[0][0], target1.values()[-20:-10])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-30:-20])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][3], target1.values()[-10:])

            np.testing.assert_almost_equal(ds[101][0], target2.values()[-40:-30])
            np.testing.assert_almost_equal(ds[101][1], cov2.values()[-60:-50])
            np.testing.assert_almost_equal(ds[101][2], self.cov_st2)
            np.testing.assert_almost_equal(ds[101][3], target2.values()[-30:-20])

            # Should also contain correct values when time-indexed with covariates not aligned
            times1 = pd.date_range(start="20090201", end="20090220", freq="D")
            times2 = pd.date_range(start="20090201", end="20090222", freq="D")
            target1 = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov1 = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )

            ds = FutureCovariatesSequentialDataset(
                target_series=[target1],
                covariates=[cov1],
                input_chunk_length=2,
                output_chunk_length=2,
            )

            np.testing.assert_almost_equal(ds[0][0], target1.values()[-4:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-4:-2])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][3], target1.values()[-2:])

            # Should fail if covariates are not long enough
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(7))

            ds = FutureCovariatesSequentialDataset(
                target_series=[target1],
                covariates=[cov1],
                input_chunk_length=2,
                output_chunk_length=2,
            )

            with self.assertRaises(ValueError):
                _ = ds[0]

        def test_dual_covariates_sequential_dataset(self):
            # Must contain (past_target, historic_future_covariates, future_covariates, future_target)

            # one target series
            ds = DualCovariatesSequentialDataset(
                target_series=self.target1,
                input_chunk_length=10,
                output_chunk_length=10,
            )
            self.assertEqual(len(ds), 81)
            self._assert_eq(
                ds[5],
                (self.target1[75:85], None, None, self.cov_st1, self.target1[85:95]),
            )

            # two target series
            ds = DualCovariatesSequentialDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=10,
                output_chunk_length=10,
            )
            self.assertEqual(len(ds), 262)
            self._assert_eq(
                ds[5],
                (self.target1[75:85], None, None, self.cov_st1, self.target1[85:95]),
            )
            self._assert_eq(
                ds[136],
                (
                    self.target2[125:135],
                    None,
                    None,
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # two target series with custom max_nr_samples
            ds = DualCovariatesSequentialDataset(
                target_series=[self.target1, self.target2],
                input_chunk_length=10,
                output_chunk_length=10,
                max_samples_per_ts=50,
            )
            self.assertEqual(len(ds), 100)
            self._assert_eq(
                ds[5],
                (self.target1[75:85], None, None, self.cov_st1, self.target1[85:95]),
            )
            self._assert_eq(
                ds[55],
                (
                    self.target2[125:135],
                    None,
                    None,
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # two targets and one covariate
            with self.assertRaises(ValueError):
                ds = DualCovariatesSequentialDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # two targets and two covariates; covariates not aligned, must contain correct values
            target1 = TimeSeries.from_values(
                np.random.randn(100)
            ).with_static_covariates(self.cov_st2_df)
            target2 = TimeSeries.from_values(
                np.random.randn(50)
            ).with_static_covariates(self.cov_st2_df)
            cov1 = TimeSeries.from_values(np.random.randn(120))
            cov2 = TimeSeries.from_values(np.random.randn(80))

            ds = DualCovariatesSequentialDataset(
                target_series=[target1, target2],
                covariates=[cov1, cov2],
                input_chunk_length=10,
                output_chunk_length=10,
            )

            np.testing.assert_almost_equal(ds[0][0], target1.values()[-20:-10])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-40:-30])
            np.testing.assert_almost_equal(ds[0][2], cov1.values()[-30:-20])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][4], target1.values()[-10:])

            np.testing.assert_almost_equal(ds[101][0], target2.values()[-40:-30])
            np.testing.assert_almost_equal(ds[101][1], cov2.values()[-70:-60])
            np.testing.assert_almost_equal(ds[101][2], cov2.values()[-60:-50])
            np.testing.assert_almost_equal(ds[101][3], self.cov_st2)
            np.testing.assert_almost_equal(ds[101][4], target2.values()[-30:-20])

            # Should also contain correct values when time-indexed with covariates not aligned
            times1 = pd.date_range(start="20090201", end="20090220", freq="D")
            times2 = pd.date_range(start="20090201", end="20090222", freq="D")
            target1 = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov1 = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )

            ds = DualCovariatesSequentialDataset(
                target_series=[target1],
                covariates=[cov1],
                input_chunk_length=2,
                output_chunk_length=2,
            )

            np.testing.assert_almost_equal(ds[0][0], target1.values()[-4:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-6:-4])
            np.testing.assert_almost_equal(ds[0][2], cov1.values()[-4:-2])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][4], target1.values()[-2:])

            # Should fail if covariates are not long enough
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(7))

            ds = DualCovariatesSequentialDataset(
                target_series=[target1],
                covariates=[cov1],
                input_chunk_length=2,
                output_chunk_length=2,
            )

            with self.assertRaises(ValueError):
                _ = ds[0]

        def test_past_covariates_shifted_dataset(self):
            # one target series
            ds = PastCovariatesShiftedDataset(
                target_series=self.target1, length=10, shift=5
            )
            self.assertEqual(len(ds), 86)
            self._assert_eq(
                ds[5], (self.target1[80:90], None, self.cov_st1, self.target1[85:95])
            )

            # two target series
            ds = PastCovariatesShiftedDataset(
                target_series=[self.target1, self.target2], length=10, shift=5
            )
            self.assertEqual(len(ds), 272)
            self._assert_eq(
                ds[5], (self.target1[80:90], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[141],
                (self.target2[130:140], None, self.cov_st2, self.target2[135:145]),
            )

            # two target series with custom max_nr_samples
            ds = PastCovariatesShiftedDataset(
                target_series=[self.target1, self.target2],
                length=10,
                shift=5,
                max_samples_per_ts=50,
            )
            self.assertEqual(len(ds), 100)
            self._assert_eq(
                ds[5], (self.target1[80:90], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[55],
                (self.target2[130:140], None, self.cov_st2, self.target2[135:145]),
            )

            # two targets and one covariate
            with self.assertRaises(ValueError):
                ds = PastCovariatesShiftedDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # two targets and two covariates
            ds = PastCovariatesShiftedDataset(
                target_series=[self.target1, self.target2],
                covariates=[self.cov1, self.cov2],
                length=10,
                shift=5,
            )
            self._assert_eq(
                ds[5],
                (
                    self.target1[80:90],
                    self.cov1[80:90],
                    self.cov_st1,
                    self.target1[85:95],
                ),
            )
            self._assert_eq(
                ds[141],
                (
                    self.target2[130:140],
                    self.cov2[130:140],
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # Should contain correct values even when covariates are not aligned
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(10))
            ds = PastCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][3], target1.values()[-3:])

            # Should also contain correct values when time-indexed with covariates not aligned
            times1 = pd.date_range(start="20090201", end="20090220", freq="D")
            times2 = pd.date_range(start="20090201", end="20090222", freq="D")
            target1 = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov1 = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )
            ds = PastCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][3], target1.values()[-3:])

            # Should fail if covariates are too short
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(5))
            ds = PastCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            with self.assertRaises(ValueError):
                _ = ds[0]

        def test_future_covariates_shifted_dataset(self):
            # one target series
            ds = FutureCovariatesShiftedDataset(
                target_series=self.target1, length=10, shift=5
            )
            self.assertEqual(len(ds), 86)
            self._assert_eq(
                ds[5], (self.target1[80:90], None, self.cov_st1, self.target1[85:95])
            )

            # two target series
            ds = FutureCovariatesShiftedDataset(
                target_series=[self.target1, self.target2], length=10, shift=5
            )
            self.assertEqual(len(ds), 272)
            self._assert_eq(
                ds[5], (self.target1[80:90], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[141],
                (self.target2[130:140], None, self.cov_st2, self.target2[135:145]),
            )

            # two target series with custom max_nr_samples
            ds = FutureCovariatesShiftedDataset(
                target_series=[self.target1, self.target2],
                length=10,
                shift=5,
                max_samples_per_ts=50,
            )
            self.assertEqual(len(ds), 100)
            self._assert_eq(
                ds[5], (self.target1[80:90], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[55],
                (self.target2[130:140], None, self.cov_st2, self.target2[135:145]),
            )

            # two targets and one covariate
            with self.assertRaises(ValueError):
                ds = FutureCovariatesShiftedDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # two targets and two covariates
            ds = FutureCovariatesShiftedDataset(
                target_series=[self.target1, self.target2],
                covariates=[self.cov1, self.cov2],
                length=10,
                shift=5,
            )
            self._assert_eq(
                ds[5],
                (
                    self.target1[80:90],
                    self.cov1[85:95],
                    self.cov_st1,
                    self.target1[85:95],
                ),
            )
            self._assert_eq(
                ds[141],
                (
                    self.target2[130:140],
                    self.cov2[135:145],
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # Should contain correct values even when covariates are not aligned
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(10))
            ds = FutureCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][3], target1.values()[-3:])

            # Should also contain correct values when time-indexed with covariates not aligned
            times1 = pd.date_range(start="20090201", end="20090220", freq="D")
            times2 = pd.date_range(start="20090201", end="20090222", freq="D")
            target1 = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov1 = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )
            ds = FutureCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][3], target1.values()[-3:])

            # Should fail if covariates are too short
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(7))
            ds = FutureCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            with self.assertRaises(ValueError):
                _ = ds[0]

        def test_dual_covariates_shifted_dataset(self):
            # one target series
            ds = DualCovariatesShiftedDataset(
                target_series=self.target1, length=10, shift=5
            )
            self.assertEqual(len(ds), 86)
            self._assert_eq(
                ds[5],
                (self.target1[80:90], None, None, self.cov_st1, self.target1[85:95]),
            )

            # two target series
            ds = DualCovariatesShiftedDataset(
                target_series=[self.target1, self.target2], length=10, shift=5
            )
            self.assertEqual(len(ds), 272)
            self._assert_eq(
                ds[5],
                (self.target1[80:90], None, None, self.cov_st1, self.target1[85:95]),
            )
            self._assert_eq(
                ds[141],
                (
                    self.target2[130:140],
                    None,
                    None,
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # two target series with custom max_nr_samples
            ds = DualCovariatesShiftedDataset(
                target_series=[self.target1, self.target2],
                length=10,
                shift=5,
                max_samples_per_ts=50,
            )
            self.assertEqual(len(ds), 100)
            self._assert_eq(
                ds[5],
                (self.target1[80:90], None, None, self.cov_st1, self.target1[85:95]),
            )
            self._assert_eq(
                ds[55],
                (
                    self.target2[130:140],
                    None,
                    None,
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # two targets and one covariate
            with self.assertRaises(ValueError):
                ds = DualCovariatesShiftedDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # two targets and two covariates
            ds = DualCovariatesShiftedDataset(
                target_series=[self.target1, self.target2],
                covariates=[self.cov1, self.cov2],
                length=10,
                shift=5,
            )
            self._assert_eq(
                ds[5],
                (
                    self.target1[80:90],
                    self.cov1[80:90],
                    self.cov1[85:95],
                    self.cov_st1,
                    self.target1[85:95],
                ),
            )
            self._assert_eq(
                ds[141],
                (
                    self.target2[130:140],
                    self.cov2[130:140],
                    self.cov2[135:145],
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

            # Should contain correct values even when covariates are not aligned
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(10))
            ds = DualCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
            np.testing.assert_almost_equal(ds[0][2], cov1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][4], target1.values()[-3:])

            # Should also contain correct values when time-indexed with covariates not aligned
            times1 = pd.date_range(start="20090201", end="20090220", freq="D")
            times2 = pd.date_range(start="20090201", end="20090222", freq="D")
            target1 = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov1 = TimeSeries.from_times_and_values(
                times2, np.random.randn(len(times2))
            )
            ds = DualCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
            np.testing.assert_almost_equal(ds[0][2], cov1.values()[-5:-2])
            np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
            np.testing.assert_almost_equal(ds[0][4], target1.values()[-3:])

            # Should fail if covariates are too short
            target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
                self.cov_st2_df
            )
            cov1 = TimeSeries.from_values(np.random.randn(7))
            ds = DualCovariatesShiftedDataset(
                target_series=[target1], covariates=[cov1], length=3, shift=2
            )
            with self.assertRaises(ValueError):
                _ = ds[0]

        def test_horizon_based_dataset(self):
            # one target series
            ds = HorizonBasedDataset(
                target_series=self.target1,
                output_chunk_length=10,
                lh=(1, 3),
                lookback=2,
            )
            self.assertEqual(len(ds), 20)
            self._assert_eq(
                ds[5], (self.target1[65:85], None, self.cov_st1, self.target1[85:95])
            )

            # two target series
            ds = HorizonBasedDataset(
                target_series=[self.target1, self.target2],
                output_chunk_length=10,
                lh=(1, 3),
                lookback=2,
            )
            self.assertEqual(len(ds), 40)
            self._assert_eq(
                ds[5], (self.target1[65:85], None, self.cov_st1, self.target1[85:95])
            )
            self._assert_eq(
                ds[25],
                (self.target2[115:135], None, self.cov_st2, self.target2[135:145]),
            )

            # two targets and one covariate
            with self.assertRaises(ValueError):
                ds = HorizonBasedDataset(
                    target_series=[self.target1, self.target2], covariates=[self.cov1]
                )

            # two targets and two covariates
            ds = HorizonBasedDataset(
                target_series=[self.target1, self.target2],
                covariates=[self.cov1, self.cov2],
                output_chunk_length=10,
                lh=(1, 3),
                lookback=2,
            )
            self._assert_eq(
                ds[5],
                (
                    self.target1[65:85],
                    self.cov1[65:85],
                    self.cov_st1,
                    self.target1[85:95],
                ),
            )
            self._assert_eq(
                ds[25],
                (
                    self.target2[115:135],
                    self.cov2[115:135],
                    self.cov_st2,
                    self.target2[135:145],
                ),
            )

        def test_get_matching_index(self):
            from darts.utils.data.utils import _get_matching_index

            # Check dividable freq
            times1 = pd.date_range(start="20100101", end="20100330", freq="D")
            times2 = pd.date_range(start="20100101", end="20100320", freq="D")
            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
            self.assertEqual(_get_matching_index(target, cov, idx=15), 5)

            # check non-dividable freq
            times1 = pd.date_range(start="20100101", end="20120101", freq="M")
            times2 = pd.date_range(start="20090101", end="20110601", freq="M")
            target = TimeSeries.from_times_and_values(
                times1, np.random.randn(len(times1))
            ).with_static_covariates(self.cov_st2_df)
            cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
            self.assertEqual(_get_matching_index(target, cov, idx=15), 15 - 7)

            # check integer-indexed series
            times2 = pd.RangeIndex(start=10, stop=90)
            target = TimeSeries.from_values(
                np.random.randn(100)
            ).with_static_covariates(self.cov_st2_df)
            cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
            self.assertEqual(_get_matching_index(target, cov, idx=15), 5)



"""
Regression ensemble model
-------------------------

An ensemble model which uses a regression model to compute the ensemble forecast.
"""
from typing import List, Optional, Sequence, Tuple, Union

from darts.logging import get_logger, raise_if, raise_if_not
from darts.models.forecasting.ensemble_model import EnsembleModel
from darts.models.forecasting.forecasting_model import (
    GlobalForecastingModel,
    LocalForecastingModel,
)
from darts.models.forecasting.linear_regression_model import LinearRegressionModel
from darts.models.forecasting.regression_model import RegressionModel
from darts.timeseries import TimeSeries
from darts.utils.utils import seq2series, series2seq

logger = get_logger(__name__)


class RegressionEnsembleModel(EnsembleModel):
    def __init__(
        self,
        forecasting_models: Union[
            List[LocalForecastingModel], List[GlobalForecastingModel]
        ],
        regression_train_n_points: int,
        regression_model=None,
    ):
        """
        Use a regression model for ensembling individual models' predictions.

        The provided regression model must implement ``fit()`` and ``predict()`` methods
        (e.g. scikit-learn regression models). Note that here the regression model is used to learn how to
        best ensemble the individual forecasting models' forecasts. It is not the same usage of regression
        as in :class:`RegressionModel`, where the regression model is used to produce forecasts based on the
        lagged series.

        Parameters
        ----------
        forecasting_models
            List of forecasting models whose predictions to ensemble
        regression_train_n_points
            The number of points to use to train the regression model
        regression_model
            Any regression model with ``predict()`` and ``fit()`` methods (e.g. from scikit-learn)
            Default: ``darts.model.LinearRegressionModel(fit_intercept=False)``
        """
        super().__init__(forecasting_models)
        if regression_model is None:
            regression_model = LinearRegressionModel(
                lags=None, lags_future_covariates=[0], fit_intercept=False
            )
        elif isinstance(regression_model, RegressionModel):
            regression_model = regression_model
        else:
            # scikit-learn like model
            regression_model = RegressionModel(
                lags_future_covariates=[0], model=regression_model
            )

        # check lags of the regression model
        raise_if_not(
            regression_model.lags == {"future": [0]},
            f"`lags` and `lags_past_covariates` of regression model must be `None`"
            f"and `lags_future_covariates` must be [0]. Given:\n"
            f"{regression_model.lags}",
        )

        self.regression_model = regression_model
        self.train_n_points = regression_train_n_points

    def _split_multi_ts_sequence(
        self, n: int, ts_sequence: Sequence[TimeSeries]
    ) -> Tuple[Sequence[TimeSeries], Sequence[TimeSeries]]:
        left = [ts[:-n] for ts in ts_sequence]
        right = [ts[-n:] for ts in ts_sequence]
        return left, right

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ):

        super().fit(
            series, past_covariates=past_covariates, future_covariates=future_covariates
        )

        # spare train_n_points points to serve as regression target
        is_single_series = isinstance(series, TimeSeries)
        if is_single_series:
            train_n_points_too_big = len(self.training_series) <= self.train_n_points
        else:
            train_n_points_too_big = any(
                [len(s) <= self.train_n_points for s in series]
            )

        raise_if(
            train_n_points_too_big,
            "regression_train_n_points parameter too big (must be smaller or "
            "equal to the number of points in training_series)",
            logger,
        )

        if is_single_series:
            forecast_training = self.training_series[: -self.train_n_points]
            regression_target = self.training_series[-self.train_n_points :]
        else:
            forecast_training, regression_target = self._split_multi_ts_sequence(
                self.train_n_points, series
            )

        for model in self.models:
            model._fit_wrapper(
                series=forecast_training,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
            )

        predictions = self._make_multiple_predictions(
            n=self.train_n_points,
            series=forecast_training,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            num_samples=1,
        )

        # train the regression model on the individual models' predictions
        self.regression_model.fit(
            series=regression_target, future_covariates=predictions
        )

        # prepare the forecasting models for further predicting by fitting them with the entire data

        # Some models (incl. Neural-Network based models) may need to be 'reset' to allow being retrained from scratch
        self.models = [model.untrained_model() for model in self.models]

        for model in self.models:
            kwargs = dict(series=series)
            if model.supports_past_covariates:
                kwargs["past_covariates"] = past_covariates
            if model.supports_future_covariates:
                kwargs["future_covariates"] = future_covariates
            model.fit(**kwargs)
        return self

    def ensemble(
        self,
        predictions: Union[TimeSeries, Sequence[TimeSeries]],
        series: Optional[Sequence[TimeSeries]] = None,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:

        is_single_series = isinstance(series, TimeSeries) or series is None
        predictions = series2seq(predictions)
        series = series2seq(series) if series is not None else [None]

        ensembled = [
            self.regression_model.predict(
                n=len(prediction), series=serie, future_covariates=prediction
            )
            for serie, prediction in zip(series, predictions)
        ]
        return seq2series(ensembled) if is_single_series else ensembled

    @property
    def extreme_lags(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
    ]:
        extreme_lags_ = super().extreme_lags
        return (extreme_lags_[0] - self.train_n_points,) + extreme_lags_[1:]



"""
NLL Exponential Scorer
----------------------

Exponential distribution negative log-likelihood Scorer.

The anomaly score is the negative log likelihood of the actual series values
under an Exponential distribution estimated from the stochastic prediction.
"""

import numpy as np
from scipy.stats import expon

from darts.ad.scorers.scorers import NLLScorer


class ExponentialNLLScorer(NLLScorer):
    def __init__(self, window: int = 1) -> None:
        super().__init__(window=window)

    def __str__(self):
        return "ExponentialNLLScorer"

    def _score_core_nllikelihood(
        self,
        deterministic_values: np.ndarray,
        probabilistic_estimations: np.ndarray,
    ) -> np.ndarray:

        # This is the ML estimate for 1/lambda, which is what scipy expects as scale.
        mu = np.mean(probabilistic_estimations, axis=1)

        # This is ML estimate for the loc - see:
        # https://github.com/scipy/scipy/blob/de80faf9d3480b9dbb9b888568b64499e0e70c19/scipy/stats/_continuous_distns.py#L1705
        loc = np.min(probabilistic_estimations, axis=1)

        return -expon.logpdf(deterministic_values, scale=mu, loc=loc)
